#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""Resegmentation"""

from typing import Optional
from typing import Type
from typing import Iterable
from typing import Dict
from typing import Text
import scipy.signal

import torch
import tempfile
import numpy as np
from .base import LabelingTask
from .base import LabelingTaskGenerator
from pyannote.audio.train.task import Task, TaskType, TaskOutput
from pyannote.core import Timeline
from pyannote.core import Annotation
from pyannote.core import SlidingWindow
from pyannote.core import SlidingWindowFeature
from pyannote.database.protocol import SpeakerDiarizationProtocol
from pyannote.database import get_annotated
from pyannote.core.utils.numpy import one_hot_decoding
from pyannote.core.utils.numpy import one_hot_encoding
from pyannote.audio.train.schedulers import ConstantScheduler
from torch.optim import SGD
from pathlib import Path
from pyannote.audio.utils.signal import Binarize

from pyannote.audio.features import FeatureExtraction
from pyannote.database import ProtocolFile
from pyannote.audio.train.model import Model
from pyannote.audio.train.model import Resolution
from pyannote.audio.train.model import Alignment
from pyannote.audio.train.task import Task


class ResegmentationGenerator(LabelingTaskGenerator):
    """Batch generator for resegmentation self-training

    Parameters
    ----------
    tasl : Task
    current_file : `ProtocolFile`
    resolution : `pyannote.core.SlidingWindow`, optional
        Override `feature_extraction.sliding_window`. This is useful for
        models that include the feature extraction step (e.g. SincNet) and
        therefore output a lower sample rate than that of the input.
    alignment : {'center', 'loose', 'strict'}, optional
        Which mode to use when cropping labels. This is useful for models that
        include the feature extraction step (e.g. SincNet) and therefore use a
        different cropping mode. Defaults to 'center'.
    duration : float, optional
        Duration of audio chunks. Defaults to 4s.
    batch_size : int, optional
        Batch size. Defaults to 32.
    allow_overlap : bool, optional
        Allow overlapping speakers. Defaults to False.
        Does not really work for now...
    mask : str, optional
        When provided, current_file[mask] is used by the loss function to weigh
        samples.
    """

    def __init__(
        self,
        task: Task,
        current_file: ProtocolFile,
        resolution: Optional[Resolution] = None,
        alignment: Optional[Alignment] = None,
        duration: float = 2,
        batch_size: int = 32,
        lock_speech: bool = False,
        allow_overlap: bool = False,
        mask: Text = None,
    ):

        self.current_file = current_file
        self.allow_overlap = allow_overlap
        self.lock_speech = lock_speech

        super().__init__(
            task,
            "@features",
            self.get_dummy_protocol(current_file),
            subset="train",
            resolution=resolution,
            alignment=alignment,
            duration=duration,
            batch_size=batch_size,
            mask=mask,
        )

    @property
    def resolution(self):
        return self.current_file["features"].sliding_window

    def get_dummy_protocol(
        self, current_file: ProtocolFile
    ) -> SpeakerDiarizationProtocol:
        """Get dummy protocol containing only `current_file`

        Parameters
        ----------
        current_file : ProtocolFile

        Returns
        -------
        protocol : SpeakerDiarizationProtocol instance
            Dummy protocol containing only `current_file` in both train,
            dev., and test sets.

        """

        class DummyProtocol(SpeakerDiarizationProtocol):
            def trn_iter(self):
                yield current_file

            def dev_iter(self):
                yield current_file

            def tst_iter(self):
                yield current_file

        return DummyProtocol()

    def postprocess_y(self, Y: np.ndarray) -> np.ndarray:
        """Generate labels for resegmentation

        Parameters
        ----------
        Y : (n_samples, n_speakers) numpy.ndarray
            Discretized annotation returned by
            `pyannote.core.utils.numpy.one_hot_encoding`.

        Returns
        -------
        y : (n_samples, 1) or (n_samples, n_speakers) numpy.ndarray
            When allow_overlap is True, y has shape (n_samples, n_speakers)
                * y[t, i] = 1 means speaker i is active
            When allow_overlap is False, y has shape (n_samples, 1)
                * y[t] = 0 indicates non-speech,
                * y[t] = i + 1 indicates speaker i.

        See also
        --------
        `pyannote.core.utils.numpy.one_hot_encoding`
        """

        # when allowing overlap, multiple speakers can be active at once.
        # hence, Y sticks to shape (n_samples, n_speakers)
        if self.allow_overlap:
            return Y

        # when overlap is not allowed, reshape Y to (n_samples, 1)
        else:

            # if speech / non-speech status is locked
            # there is no class for non-speech.
            if self.lock_speech:
                y = np.argmax(Y, axis=1)
                return np.int64(y)[:, np.newaxis]

            # otherwise, we add a non-speech class at index 0
            else:

                # +1 because...
                y = np.argmax(Y, axis=1) + 1

                # ... 0 is for non-speech
                non_speech = np.sum(Y, axis=1) == 0
                y[non_speech] = 0

                return np.int64(y)[:, np.newaxis]

    @property
    def specifications(self) -> Dict:
        """Task & sample specifications

        Returns
        -------
        specs : `dict`
            ['task'] (`pyannote.audio.train.Task`) : task
            ['X']['dimension'] (`int`) : features dimension
            ['y']['classes'] (`list`) : list of classes
        """

        specs = {
            "X": {"dimension": self.current_file["features"].dimension},
            "task": self.task,
        }

        if self.allow_overlap:
            specs["y"] = {"classes": self.segment_labels_}

        else:
            if self.lock_speech:
                # when locking speech / non-speech status,
                # there is no non-speech class
                specs["y"] = {"classes": self.segment_labels_}
            else:
                # when speech / non-speech status can be updated,
                # one must add a non-speech class
                specs["y"] = {"classes": ["non_speech"] + self.segment_labels_}

        return specs


class Resegmentation(LabelingTask):
    """Re-segmentation

    Parameters
    ----------
    feature_extraction : FeatureExtraction
        Feature extraction.
    Architecture : Model subclass
    architecture_params : dict
    epochs : `int`, optional
        (Self-)train for that many epochs. Defaults to 5.
    ensemble : `int`, optional
        Average output of last `ensemble` epochs. Defaults to no ensembling.
    duration : float, optional
        Duration of audio chunks. Defaults to 2s.
    step : `float`, optional
        Ratio of audio chunk duration used as step between two consecutive
        audio chunks. Defaults to 0.1.
    batch_size : int, optional
        Batch size. Defaults to 32.
    device : `torch.device`, optional
    lock_speech: `boolean`, optional
        Keep speech/non-speech state unchanged. Defaults to False.
    allow_overlap : bool, optional
        Allow overlapping speakers. Defaults to False.
    mask : str, optional
        When provided, current_file[mask] is used by the loss function to weigh
        samples.
    """

    def __init__(
        self,
        feature_extraction: FeatureExtraction,
        Architecture: Type[Model],
        architecture_params: dict,
        lock_speech: bool = False,
        epochs: int = 5,
        learning_rate: float = 0.1,
        ensemble: int = 1,
        duration: float = 2.0,
        step: float = 0.1,
        n_jobs: int = 1,
        device: torch.device = None,
        batch_size: int = 32,
        allow_overlap: bool = False,
        mask: Text = None,
    ):

        self.feature_extraction = feature_extraction

        self.Architecture = Architecture
        self.architecture_params = architecture_params

        self.epochs = epochs
        self.learning_rate = learning_rate

        self.ensemble = ensemble

        self.n_jobs = n_jobs

        self.lock_speech = lock_speech
        self.allow_overlap = allow_overlap
        self.mask = mask

        if device is None:
            device = "cuda" if torch.cuda.is_available() else "cpu"
        self.device_ = torch.device(device)

        super().__init__(
            duration=duration, batch_size=batch_size, per_epoch=None, step=step
        )

    @property
    def task(self):
        if self.allow_overlap:
            # when allowing overlap, multiple speakers can be active at the
            # same time (hence multi-label classification task)
            return Task(
                type=TaskType.MULTI_LABEL_CLASSIFICATION, output=TaskOutput.SEQUENCE
            )

        else:
            # when overlap is not allowed, only one speaker can be active at
            # a particular time (hence: multi-class classification)
            return Task(
                type=TaskType.MULTI_CLASS_CLASSIFICATION, output=TaskOutput.SEQUENCE
            )

    def get_batch_generator(
        self, current_file: ProtocolFile
    ) -> ResegmentationGenerator:
        """Get batch generator for current file

        Parameters
        ----------
        current_file : `dict`
            Dictionary obtained by iterating over a subset of a
            `pyannote.database.Protocol` instance.

        Returns
        -------
        batch_generator : `ResegmentationGenerator`
        """

        resolution = self.Architecture.get_resolution(
            self.task, **self.architecture_params
        )
        alignment = self.Architecture.get_alignment(
            self.task, **self.architecture_params
        )

        return ResegmentationGenerator(
            self.task,
            current_file,
            resolution=resolution,
            alignment=alignment,
            duration=self.duration,
            batch_size=self.batch_size,
            lock_speech=self.lock_speech,
            allow_overlap=self.allow_overlap,
            mask=self.mask,
        )

    def _decode(
        self,
        current_file: ProtocolFile,
        hypothesis: Annotation,
        scores: SlidingWindowFeature,
        labels: Iterable,
    ) -> Annotation:

        N, K = scores.data.shape

        if self.allow_overlap:
            active_speakers = scores.data > 0.5

        else:
            if self.lock_speech:
                active_speakers = np.argmax(scores.data, axis=1) + 1

            else:
                active_speakers = np.argmax(scores.data, axis=1)

        # reconstruct annotation
        new_hypothesis = one_hot_decoding(
            active_speakers, scores.sliding_window, labels=labels
        )

        new_hypothesis.uri = hypothesis.uri

        if self.lock_speech:
            speech = hypothesis.get_timeline().support()
            new_hypothesis = new_hypothesis.crop(speech)

        return new_hypothesis

    def __call__(
        self,
        current_file: ProtocolFile,
        hypothesis: Annotation,
        debugging: bool = False,
    ) -> Annotation:
        """Apply resegmentation using self-supervised sequence labeling

        Parameters
        ----------
        current_file : ProtocolFile
            Dictionary obtained by iterating over a subset of a
            `pyannote.database.Protocol` instance.
        hypothesis : Annotation, optional
            Current diarization output. Defaults to current_file['hypothesis'].

        Returns
        -------
        new_hypothesis : Annotation
            Updated diarization output.
        """

        # make sure current_file is not modified
        current_file = dict(current_file)

        current_file["annotation"] = hypothesis
        current_file["features"] = self.feature_extraction(current_file)

        debug = {}

        # when locking speech / non-speech status, we add a (or update
        # existing) mask so that the loss is not computed on non-speech regions
        if self.lock_speech:

            encoded = one_hot_encoding(
                hypothesis,
                get_annotated(current_file),
                current_file["features"].sliding_window,
                mode="center",
            )
            speech = 1.0 * (np.sum(encoded, axis=1, keepdims=True) > 0)
            current_file["speech"] = speech
            debug["speech"] = speech

            if self.mask is None:
                self.mask = "speech"

            else:
                mask = current_file[self.mask]
                current_file[self.mask] = mask * speech.align(mask)

            debug["mask"] = current_file[self.mask]

        batch_generator = self.get_batch_generator(current_file)

        model = self.Architecture(
            batch_generator.specifications, **self.architecture_params
        )

        chunks = SlidingWindow(duration=self.duration, step=self.step * self.duration)

        # create a temporary directory to store models and log files
        # it is removed automatically before returning.
        with tempfile.TemporaryDirectory() as train_dir:

            epochs = self.fit_iter(
                model,
                batch_generator,
                warm_start=0,
                epochs=self.epochs,
                get_optimizer=SGD,
                scheduler=ConstantScheduler(),
                learning_rate=self.learning_rate,
                train_dir=Path(train_dir),
                verbosity=1,
                device=self.device,
                callbacks=None,
                n_jobs=self.n_jobs,
            )

            scores = []
            for i, current_model in enumerate(epochs):

                # do not compute scores that are not used in later ensembling
                # simply jump to next training epoch (except when debugging)
                if not debugging and i < self.epochs - self.ensemble:
                    continue

                current_model.eval()

                scores.append(
                    current_model.slide(
                        current_file["features"],
                        chunks,
                        batch_size=self.batch_size,
                        device=self.device,
                        return_intermediate=None,
                        progress_hook=None,
                    )
                )
                current_model.train()

        debug["scores"] = scores

        # ensemble scores
        scores = SlidingWindowFeature(
            np.mean([s.data for s in scores[-self.ensemble :]], axis=0),
            scores[-1].sliding_window,
        )
        debug["final_scores"] = scores

        labels = batch_generator.specifications["y"]["classes"]
        if not self.lock_speech:
            labels = labels[1:]

        debug["labels"] = labels

        decoded = self._decode(current_file, hypothesis, scores, labels)

        decoded.debug = debug
        return decoded


class ResegmentationWithOverlap(Resegmentation):
    """Re-segmentation with overlap

    Parameters
    ----------
    feature_extraction : FeatureExtraction
        Feature extraction.
    Architecture : Model subclass
    architecture_params : dict
    overlap_threshold : `float`, optional
        Defaults to 0.5.
    lock_speech: `boolean`, optional
        Keep speech/non-speech state unchanged. Defaults to False.
    epochs : `int`, optional
        (Self-)train for that many epochs. Defaults to 5.
    ensemble : `int`, optional
        Average output of last `ensemble` epochs. Defaults to no ensembling.
    duration : float, optional
        Duration of audio chunks. Defaults to 2s.
    step : `float`, optional
        Ratio of audio chunk duration used as step between two consecutive
        audio chunks. Defaults to 0.1.
    batch_size : int, optional
        Batch size. Defaults to 32.
    device : `torch.device`, optional
    mask : str, optional
        When provided, current_file[mask] is used by the loss function to weigh
        samples.
    """

    def __init__(
        self,
        feature_extraction: FeatureExtraction,
        Architecture: Type[Model],
        architecture_params: dict,
        lock_speech: bool = False,
        overlap_threshold: float = 0.5,
        epochs: int = 5,
        learning_rate: float = 0.1,
        ensemble: int = 1,
        duration: float = 2.0,
        step: float = 0.1,
        n_jobs: int = 1,
        device: torch.device = None,
        batch_size: int = 32,
        mask: Text = None,
    ):

        super().__init__(
            feature_extraction,
            Architecture,
            architecture_params,
            lock_speech=lock_speech,
            epochs=epochs,
            learning_rate=learning_rate,
            ensemble=ensemble,
            duration=duration,
            step=step,
            n_jobs=n_jobs,
            device=device,
            batch_size=batch_size,
            mask=mask,
        )

        self.overlap_threshold = overlap_threshold
        self.binarizer_ = Binarize(
            onset=self.overlap_threshold,
            offset=self.overlap_threshold,
            scale="absolute",
            log_scale=True,
        )

    def _decode(
        self,
        current_file: ProtocolFile,
        hypothesis: Annotation,
        scores: SlidingWindowFeature,
        labels: Iterable,
    ) -> Annotation:

        # obtain overlapped speech regions
        overlap = self.binarizer_.apply(current_file["overlap"], dimension=1)

        frames = scores.sliding_window
        N, K = scores.data.shape

        if self.lock_speech:

            # K = 1 <~~> only non-speech
            # K = 2 <~~> just one speaker
            if K < 3:
                return hypothesis

            # sequence of two most likely speaker indices
            # (even when non-speech is in fact the most likely class)
            best_speakers_indices = np.argsort(-scores.data[:, 1:], axis=1)[:, :2]

            active_speakers = np.zeros((N, K - 1), dtype=np.int64)

            # start by assigning most likely speaker...
            for t, k in enumerate(best_speakers_indices[:, 0]):
                active_speakers[t, k] = 1

            # ... then add second most likely speaker in overlap regions
            T = frames.crop(overlap, mode="strict")

            # because overlap may use a different feature extraction step
            # it might happen that T contains indices slightly large than
            # the actual number of frames. the line below remove any such
            # indices.
            T = T[T < N]

            # mark second most likely speaker as active
            active_speakers[T, best_speakers_indices[T, 1]] = 1

            # reconstruct annotation
            new_hypothesis = one_hot_decoding(active_speakers, frames, labels=labels)

            # revert non-speech regions back to original
            speech = hypothesis.get_timeline().support()
            new_hypothesis = new_hypothesis.crop(speech)

        else:

            # K = 1 <~~> only non-speech
            if K < 2:
                return hypothesis

            # sequence of two most likely class indices
            # sequence of two most likely class indices
            # (including 0=non-speech)
            best_speakers_indices = np.argsort(-scores.data, axis=1)[:, :2]

            active_speakers = np.zeros((N, K - 1), dtype=np.int64)

            # start by assigning the most likely speaker...
            for t, k in enumerate(best_speakers_indices[:, 0]):
                # k = 0 is for non-speech
                if k > 0:
                    active_speakers[t, k - 1] = 1

            # ... then add second most likely speaker in overlap regions
            T = frames.crop(overlap, mode="strict")

            # because overlap may use a different feature extraction step
            # it might happen that T contains indices slightly large than
            # the actual number of frames. the line below remove any such
            # indices.
            T = T[T < N]

            # remove timesteps where second most likely class is non-speech
            T = T[best_speakers_indices[T, 1] > 0]

            # mark second most likely speaker as active
            active_speakers[T, best_speakers_indices[T, 1] - 1] = 1

            # reconstruct annotation
            new_hypothesis = one_hot_decoding(active_speakers, frames, labels=labels)

        new_hypothesis.uri = hypothesis.uri
        return new_hypothesis



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

from pathlib import Path
from pyannote.database import ProtocolFile
from pyannote.core import Annotation
from pyannote.database.util import load_rttm


class RemoveNonSpeech:
    """Remove (precomputed) non-speech regions from annotation

    This is useful for speaker verification databases where files contain only
    one speaker but actual speech regions are not annotated.

    Parameters
    ----------
    sad_rttm : Path
        Path to RTTM file containing speech activity detection result.

    Usage
    -----
    1. Prepare a file containing speech activity detection results in RTTM
       format (here /path/to/sad.rttm)
    2. Add a "preprocessors" section in your "config.yml" configuration file
        --------------------------------------------------------------------
        preprocessors:
            annotation:
                name: pyannote.audio.preprocessors.RemoveNonSpeech
                params:
                    sad_rttm: /path/to/sad.rttm
        --------------------------------------------------------------------
    3. Enjoy your updated "annotation" key (where non-speech regions are
       removed).
    """

    def __init__(self, sad_rttm: Path = None):
        self.sad_rttm = sad_rttm
        self.sad_ = load_rttm(self.sad_rttm)

    def __call__(self, current_file: ProtocolFile) -> Annotation:

        # get speech regions as Annotation instances
        speech_regions = self.sad_.get(current_file["uri"], Annotation())

        # remove non-speech regions from current annotation
        # aka only keep speech regions
        try:
            annotation = current_file["annotation"]
            return annotation.crop(speech_regions.get_timeline())

        # this haapens when current_file has no "annotation" key
        # (e.g. for file1 and file2 in speaker verification trials)
        # in that case, we return speech regions directly
        except KeyError:
            return speech_regions



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


import yaml
import io
from pathlib import Path
from glob import glob
import numpy as np
from numpy.lib.format import open_memmap

from pyannote.core import SlidingWindow, SlidingWindowFeature
from pyannote.database.util import get_unique_identifier
from pyannote.audio.utils.path import mkdir_p


class PyannoteFeatureExtractionError(Exception):
    pass


class Precomputed:
    """Precomputed features

    Parameters
    ----------
    root_dir : `str`
        Path to directory where precomputed features are stored.
    use_memmap : `bool`, optional
        Defaults to False.
    sliding_window : `SlidingWindow`, optional
        Sliding window used for feature extraction. This is not used when
        `root_dir` already exists and contains `metadata.yml`.
    dimension : `int`, optional
        Dimension of feature vectors. This is not used when `root_dir` already
        exists and contains `metadata.yml`.
    classes : iterable, optional
        Human-readable name for each dimension.

    Notes
    -----
    If `root_dir` directory does not exist, one must provide both
    `sliding_window` and `dimension` parameters in order to create and
    populate file `root_dir/metadata.yml` when instantiating.

    """

    def get_path(self, item):
        uri = get_unique_identifier(item)
        path = "{root_dir}/{uri}.npy".format(root_dir=self.root_dir, uri=uri)
        return path

    def __init__(
        self,
        root_dir=None,
        use_memmap=False,
        sliding_window=None,
        dimension=None,
        classes=None,
        augmentation=None,
    ):

        if augmentation is not None:
            msg = "Data augmentation is not supported by `Precomputed`."
            raise ValueError(msg)

        super(Precomputed, self).__init__()
        self.root_dir = Path(root_dir).expanduser().resolve(strict=False)
        self.use_memmap = use_memmap

        path = self.root_dir / "metadata.yml"
        if path.exists():

            with io.open(path, "r") as f:
                params = yaml.load(f, Loader=yaml.SafeLoader)

            self.dimension_ = params.pop("dimension")
            self.classes_ = params.pop("classes", None)
            self.sliding_window_ = SlidingWindow(**params)

            if dimension is not None and self.dimension_ != dimension:
                msg = 'inconsistent "dimension" (is: {0}, should be: {1})'
                raise ValueError(msg.format(dimension, self.dimension_))

            if classes is not None and self.classes_ != classes:
                msg = 'inconsistent "classes" (is {0}, should be: {1})'
                raise ValueError(msg.format(classes, self.classes_))

            if (sliding_window is not None) and (
                (sliding_window.start != self.sliding_window_.start)
                or (sliding_window.duration != self.sliding_window_.duration)
                or (sliding_window.step != self.sliding_window_.step)
            ):
                msg = 'inconsistent "sliding_window"'
                raise ValueError(msg)

        else:

            if dimension is None:
                if classes is None:
                    msg = (
                        f"Please provide either `dimension` or `classes` "
                        f"parameters (or both) when instantiating "
                        f"`Precomputed`."
                    )
                dimension = len(classes)

            if sliding_window is None or dimension is None:
                msg = (
                    f"Either directory {self.root_dir} does not exist or it "
                    f"does not contain precomputed features. In case it exists "
                    f"and this was done on purpose, please provide both "
                    f"`sliding_window` and `dimension` parameters when "
                    f"instantianting `Precomputed`."
                )
                raise ValueError(msg)

            # create parent directory
            mkdir_p(path.parent)

            params = {
                "start": sliding_window.start,
                "duration": sliding_window.duration,
                "step": sliding_window.step,
                "dimension": dimension,
            }
            if classes is not None:
                params["classes"] = classes

            with io.open(path, "w") as f:
                yaml.dump(params, f, default_flow_style=False)

            self.sliding_window_ = sliding_window
            self.dimension_ = dimension
            self.classes_ = classes

    def augmentation():
        doc = "Data augmentation."

        def fget(self):
            return None

        def fset(self, augmentation):
            if augmentation is not None:
                msg = "Data augmentation is not supported by `Precomputed`."
                raise AttributeError(msg)

        return locals()

    augmentation = property(**augmentation())

    @property
    def sliding_window(self):
        """Sliding window used for feature extraction"""
        return self.sliding_window_

    @property
    def dimension(self):
        """Dimension of feature vectors"""
        return self.dimension_

    @property
    def classes(self):
        """Human-readable label of each dimension"""
        return self.classes_

    def __call__(self, current_file):
        """Obtain features for file

        Parameters
        ----------
        current_file : dict
            `pyannote.database` files.

        Returns
        -------
        features : `pyannote.core.SlidingWindowFeature`
            Features
        """

        path = Path(self.get_path(current_file))

        if not path.exists():
            uri = current_file["uri"]
            database = current_file["database"]
            msg = (
                f"Directory {self.root_dir} does not contain "
                f'precomputed features for file "{uri}" of '
                f'"{database}" database.'
            )
            raise PyannoteFeatureExtractionError(msg)

        if self.use_memmap:
            data = np.load(str(path), mmap_mode="r")
        else:
            data = np.load(str(path))

        return SlidingWindowFeature(data, self.sliding_window_)

    def crop(self, current_file, segment, mode="center", fixed=None):
        """Fast version of self(current_file).crop(segment, **kwargs)

        Parameters
        ----------
        current_file : dict
            `pyannote.database` file.
        segment : `pyannote.core.Segment`
            Segment from which to extract features.

        Returns
        -------
        features : (n_frames, dimension) numpy array
            Extracted features

        See also
        --------
        `pyannote.core.SlidingWindowFeature.crop`
        """

        # match default FeatureExtraction.crop behavior
        if mode == "center" and fixed is None:
            fixed = segment.duration

        memmap = open_memmap(self.get_path(current_file), mode="r")
        swf = SlidingWindowFeature(memmap, self.sliding_window_)
        result = swf.crop(segment, mode=mode, fixed=fixed)
        del memmap
        return result

    def shape(self, item):
        """Faster version of precomputed(item).data.shape"""
        memmap = open_memmap(self.get_path(item), mode="r")
        shape = memmap.shape
        del memmap
        return shape

    def dump(self, item, features):
        path = Path(self.get_path(item))
        mkdir_p(path.parent)
        np.save(path, features.data)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2019-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""
Background generators
=====================

This module provides ways to send a data generator to one (or multiple)
background thread(s).

A typical use case is when training a neural network using mini-batches:
batches are produced on CPU and consumed (often faster) on GPU, leading to
sub-optimal usage of GPU resource (hence slow training).

`BackgroundGenerator` allows to send the CPU producer to a background thread so
that it can generate a new batch while the previous one is being consumed by
the GPU.

`AdaptiveBackgroundGenerator` goes one step further and uses a pool of
background threads whose size automatically (and continuously) adapts the
production rate to the consumption rate.
"""

import threading
import collections
import queue
import time
from typing import Iterator, Callable
import numpy as np


class BackgroundGenerator(threading.Thread):
    """Background generator with production/consumption time estimates

    Parameters
    ----------
    producer: generator function
        Generator function that takes no argument and yield (a possibly
        infinite number of) samples. This would typically be a BatchGenerator
        instance but can be any function that "yields" samples.
    prefetch: int, optional
        Maximum number of samples that can be prefetched and stored in a queue.
        Defaults to 1. In case the consumer is slower than the producer and the
        queue is full, the producer is paused until one sample is consumed.

    Usage
    -----
    >>> import time

    # a dummy producer that yield 'sample' string every 10ms.
    >>> def produce():
    ...     while True:
    ...        time.sleep(0.010)
    ...        yield 'sample'

    # a dummy consumer that takes 1ms to consume a sample
    >>> def consume(sample):
    ...     time.sleep(0.001)

    # create background generator from producer
    >>> generator = BackgroundGenerator(produce)

    # produce and consume 100 samples
    >>> for i in range(100):
    ...     sample = next(generator)
    ...     consume(sample)

    >>> p = generator.production_time
    >>> print(f'Production time estimate: {1000 * p:.0f}ms')
    # Production time estimate: 10ms

    >>> c = generator.consumption_time
    >>> print(f'Consumption time estimate: {1000 * c:.0f}ms')
    # Consumption time estimate: 1ms

    # kill background generator (and associated thread)
    >>> generator.deactivate()
    >>> sample = next(generator)
    # StopIteration: Background generator is no longer active.
    """

    def __init__(self, producer: Callable[[], Iterator], prefetch: int = 1):
        super().__init__(daemon=True)
        self.producer = producer
        self.prefetch = prefetch

        self.activated_ = True
        self.producer_ = producer()

        # used to keep track of how long it took to generate latest samples
        self.production_time_ = collections.deque([], max(10, 2 * self.prefetch))

        # used to keep track of how long it took to consume latest samples
        self.consumption_time_ = collections.deque([], max(10, 2 * self.prefetch))

        # used to keep track of last time
        self.last_ready_ = None

        # queue meant to store at most 'self.prefetch' prefetched samples
        self.queue_ = queue.Queue(self.prefetch)

        # start generator in a new thread
        self.start()

    def reset(self) -> None:
        """Reset production and consumption time estimators"""
        self.production_time_.clear()
        self.consumption_time_.clear()

    def deactivate(self) -> None:
        """Stop background generator"""
        self.activated_ = False
        # unlock queue stuck at line queue.put() in self.run()
        _ = self.queue_.get()

    @property
    def production_time(self) -> float:
        """Estimated time needed by the generator to yield a sample.

        This is computed as the median production time of the last few samples.

        Returns
        -------
        production_time : float or np.NAN
            Estimated time needed by the generator to yield a new sample, in
            seconds. Until enough samples have been yielded to accurately
            estimate production time, it is set to np.NAN.
        """

        if len(self.production_time_) < max(10, 2 * self.prefetch):
            return np.NAN
        return np.median(self.production_time_)

    @property
    def consumption_time(self) -> float:
        """Estimated time needed by the consumer to process a sample

        This is computed as the median consumption time of the last few samples.

        Returns
        -------
        consumption_time : float or np.NAN
            Estimated time needed by the consumer to process a sample, in
            seconds. Until enough samples have been consumed to accurately
            estimate consumption time, it is set to np.NAN.
        """
        if len(self.consumption_time_) < max(10, 2 * self.prefetch):
            return np.NAN
        return np.median(self.consumption_time_)

    def run(self) -> None:
        """Called by self.start(), should not be called directly."""

        # keep going until the background generator is deactivated
        while self.activated_:

            # produce a new sample
            _t = time.time()
            try:
                sample = next(self.producer_)
            except StopIteration:
                sample = None

            # keep track of how long it took to produce
            self.production_time_.append(time.time() - _t)

            # put the new sample into the queue for later consumption.
            # note that this line is blocking when the queue is full.
            # calling self.queue_.get() in self.__next__() or self.deactivate()
            # will eventually unblock it.
            self.queue_.put(sample)

    def __next__(self):
        """Produce new sample"""

        # raise a StopIteration once the generator has been deactivated
        if not self.activated_:
            msg = "Background generator is no longer active."
            raise StopIteration(msg)

        # keep track of how long it took to consume the last sample
        t = time.time()
        if self.last_ready_ is not None:
            self.consumption_time_.append(t - self.last_ready_)

        # get a new sample from the queue
        sample = self.queue_.get()

        # this happens when producer stopped yielding samples
        if sample is None:
            msg = "Producer stopped yielding samples."
            raise StopIteration(msg)

        # keep track of the last time a sample was taken from the queue
        self.last_ready_ = time.time()

        # actually return the new sample
        return sample

    def __iter__(self):
        return self


class AdaptiveBackgroundGenerator:
    """Adaptive pool of background generators

    The pool is initialized with only one background generator.

    Once production and consumption time estimates are available (after a short
    warm-up period of time), the pool will incrementally adapt the number of
    background generators to ensure that it produces samples fast enough for
    the consumer.

    Parameters
    ----------
    producer: generator function
        Generator function that takes no argument and yield (a possibly
        infinite number of) samples. This would typically be a BatchGenerator
        instance but can be any function that "yields" samples.
    n_jobs : int, optional
        Maximum number of background generators that can be created to keep up
        with consumer. Defaults to 4.
    prefetch : int, optional
        Maximum number of samples that can be prefetched by each background
        generator. See BackgroundGenerator documentation for more details.
        Defaults to 10.
    verbose : bool, optional
        Print a message when a background generator is added to (or removed
        from) the pool.

    Usage
    -----
    >>> import time

    # A producer that takes 5ms to produce a new sample
    >>> def producer():
    ...     while True:
    ...         time.sleep(0.005)
    ...         yield 'data'

    # A slow consumer that takes 5ms to consume a sample
    >>> def slow_consumer(data): time.sleep(0.005)

    # A fast consumer that takes 1ms to consume a sample
    >>> def fast_consumer(data): time.sleep(0.001)

    # send producer to the background and allows for at most 6 threads
    >>> generator = AdaptiveBackgroundGenerator(producer, n_jobs=6)

    >>> for _ in range(1000): fast_consumer(next(generator))
    >>> print(f'When consumer is fast, generator uses {len(generator)} thread(s).')
    # prints: "When consumer is fast, generator uses 4 thread(s)."

    >>> for _ in range(1000): slow_consumer(next(generator))
    >>> print(f'When consumer is slow, generator uses {len(generator)} thread(s).')
    # prints: "When consumer is slow, generator uses 1 thread(s)."

    # deactivate generator (and stop background threads)
    >>> generator.deactivate()
    >>> _ = next(generator)
    # raises: "StopIteration: Background generator is no longer active."
    """

    def __init__(
        self,
        producer: Callable[[], Iterator],
        n_jobs: int = 4,
        prefetch: int = 10,
        verbose: bool = False,
    ):

        self.producer = producer
        self.n_jobs = n_jobs
        self.prefetch = prefetch
        self.verbose = verbose

        # current pool of active background generators
        self.generators_ = []

        if self.verbose:
            msg = f"Starting with one producer."
            print(msg)

        # start by creating one background generator to the pool
        self._add_generator()

        # initialize main sample generator (used in __next__)
        self.samples_ = self._sample()

        # set to True once the maximum number of generators is reached.
        # (used to avoid repeating a message when verbose is True)
        self.reached_max_ = False

    def deactivate(self) -> None:
        """Stop background generator"""
        n_jobs = len(self.generators_)
        for _ in range(n_jobs):
            self._remove_generator()

    def _add_generator(self) -> None:
        """Add one more producer to the pool"""

        self.generators_.append(
            BackgroundGenerator(self.producer, prefetch=self.prefetch)
        )

        for g in self.generators_:
            g.reset()

    def _remove_generator(self, index: int = None) -> None:
        """Remove one producer from the pool

        Parameters
        ----------
        index : int, optional
            When provided, remove `index`th producer.
            Defaults to removing the last producer.
        """

        if index is None:
            n_jobs = len(self.generators_)
            index = n_jobs - 1

        g = self.generators_.pop(index)
        g.deactivate()

        for g in self.generators_:
            g.reset()

        self.reached_max_ = False

    def __len__(self):
        """Return current number of producers"""
        return len(self.generators_)

    @property
    def consumption_time(self) -> float:
        """Estimated time needed by the consumer to process a sample

        This is computed as the average of estimated consumption times of all
        currently active background generators.

        Returns
        -------
        consumption_time : float or np.NAN
            Estimated time needed by the consumer to process a sample, in
            seconds. Until enough samples have been consumed to accurately
            estimate consumption time, it is set to np.NAN.
        """

        # corner case when generator has been deactivated
        if not self.generators_:
            return np.NAN

        return np.mean([g.consumption_time for g in self.generators_])

    @property
    def production_time(self) -> float:
        """Estimated time needed by the generator to yield a sample.

        This is computed as the average estimated production time of all
        currently active background generators.

        Returns
        -------
        production_time : float or np.NAN
            Estimated time needed by the generator to yield a new sample, in
            seconds. Until enough samples have been yielded to accurately
            estimate production time, it is set to np.NAN.
        """

        # corner case when generator has been deactivated
        if not self.generators_:
            return np.NAN

        return np.mean([g.production_time for g in self.generators_])

    def __iter__(self):
        return self

    def __next__(self):
        return next(self.samples_)

    def _sample(self) -> Iterator:
        """Iterate over (and manage) pool of generators"""

        # loop forever
        while True:

            if not self.generators_:
                msg = "Background generator is no longer active."
                raise StopIteration(msg)

            dead_generators = []
            for index, g in enumerate(self.generators_):

                try:
                    sample = next(g)
                except StopIteration:
                    # mark this generator as dead
                    dead_generators.append(index)
                    continue

                yield sample

            if self.verbose and dead_generators:
                msg = f"Replacing {len(dead_generators)} exhausted producers."
                print(msg)

            # replace dead generators by new ones
            for index in reversed(dead_generators):
                self._remove_generator(index=index)
            for _ in dead_generators:
                self._add_generator()

            consumption_time = self.consumption_time
            production_time = self.production_time

            if np.isnan(consumption_time) or np.isnan(production_time):
                continue

            n_jobs = len(self.generators_)
            ratio = production_time / consumption_time

            # consumption_time < production_time
            if ratio > 1:
                if n_jobs < self.n_jobs:

                    if self.verbose:
                        msg = (
                            f"Adding one producer because consumer is "
                            f"{ratio:.2f}x faster than current {n_jobs:d} "
                            f"producer(s)."
                        )
                        print(msg)

                    self._add_generator()

                else:
                    if not self.reached_max_ and self.verbose:
                        msg = (
                            f"Consumer is {ratio:.2f}x faster than the pool of "
                            f"{n_jobs:d} producer(s) but the maximum number of "
                            f"producers has been reached."
                        )
                        print(msg)

                    self.reached_max_ = True

            # production_time < consumption_time * (n_jobs - 1) / n_jobs
            elif (ratio < (n_jobs - 1) / n_jobs) and n_jobs > 1:

                if self.verbose:
                    msg = (
                        f"Removing one producer because consumer is "
                        f"{1 / ratio:.2f}x slower than current {n_jobs:d} "
                        f"producer(s)."
                    )
                    print(msg)

                self._remove_generator()



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr
# Juan Manuel CORIA

import torch
from .base import RepresentationLearning


class ContrastiveLoss(RepresentationLearning):
    """Contrasitve loss

    TODO explain

    Parameters
    ----------
    duration : float, optional
        Chunks duration, in seconds. Defaults to 1.
    min_duration : float, optional
        When provided, use chunks of random duration between `min_duration` and
        `duration` for training. Defaults to using fixed duration chunks.
    per_turn : int, optional
        Number of chunks per speech turn. Defaults to 1.
        If per_turn is greater than one, embeddings of the same speech turn
        are averaged before comparison. The intuition is that it might help
        learn embeddings meant to be averaged/summed.
    per_label : `int`, optional
        Number of sequences per speaker in each batch. Defaults to 1.
    per_fold : `int`, optional
        Number of different speakers per batch. Defaults to 32.
    per_epoch : `float`, optional
        Force total audio duration per epoch, in days.
        Defaults to total duration of protocol subset.
    label_min_duration : `float`, optional
        Remove speakers with less than that many seconds of speech.
        Defaults to 0 (i.e. keep them all).
    metric : {'euclidean', 'cosine', 'angular'}, optional
        Defaults to 'cosine'.
    margin: float, optional
        Margin multiplicative factor. Defaults to 0.2.

    Reference
    ---------
    TODO
    """

    def __init__(
        self,
        duration: float = 1.0,
        min_duration: float = None,
        per_turn: int = 1,
        per_label: int = 1,
        per_fold: int = 32,
        per_epoch: float = None,
        label_min_duration: float = 0.0,
        # FIXME create a Literal type for metric
        # FIXME maybe in pyannote.core.utils.distance
        metric: str = "cosine",
        # FIXME homogeneize the meaning of margin parameter
        # FIXME it has a different meaning in ArcFace, right?
        margin: float = 0.2,
    ):

        super().__init__(
            duration=duration,
            min_duration=min_duration,
            per_turn=per_turn,
            per_label=per_label,
            per_fold=per_fold,
            per_epoch=per_epoch,
            label_min_duration=label_min_duration,
        )

        self.metric = metric
        self.margin = margin
        # FIXME see above
        self.margin_ = self.margin * self.max_distance

    def batch_loss(self, batch):
        """Compute loss for current `batch`

        Parameters
        ----------
        batch : `dict`
            ['X'] (`numpy.ndarray`)
            ['y'] (`numpy.ndarray`)

        Returns
        -------
        batch_loss : `dict`
            ['loss'] (`torch.Tensor`) : Triplet loss
        """

        fX, y = self.embed(batch)

        # calculate the distances between every sample in the batch
        batch_size = fX.size(0)
        dist = self.pdist(fX).to(self.device_)

        # calculate the ground truth for each pair
        # TODO. this can be done much more cleanly with
        # pyannote.core.utils.distance.pdist(y, metric='equal')
        gt = []
        for i in range(batch_size - 1):
            for j in range(i + 1, batch_size):
                gt.append(int(y[i] != y[j]))
        gt = torch.Tensor(gt).float().to(self.device_)

        # Calculate the losses as described in the paper
        losses = (1 - gt) * torch.pow(dist, 2) + gt * torch.pow(
            torch.clamp(self.margin_ - dist, min=1e-8), 2
        )

        # FIXME: why divive by 2?
        losses = torch.sum(losses) / 2
        # Average by batch size if requested
        # FIXME: switch to torch.mean directly (size_average has been removed)
        loss = losses / dist.size(0)

        return {"loss": loss, "loss_contrastive": loss}



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2019-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


"""
Neural building blocks for speaker diarization
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Usage:
  pyannote-audio (sad | scd | ovl | emb | dom) train    [--cpu | --gpu] [options] <root>     <protocol>
  pyannote-audio (sad | scd | ovl | emb | dom) validate [--cpu | --gpu] [options] <train>    <protocol>
  pyannote-audio (sad | scd | ovl | emb | dom) apply    [--cpu | --gpu] [options] <validate> <protocol>
  pyannote-audio -h | --help
  pyannote-audio --version

This command line tool can be used to train, validate, and apply neural networks
for the following blocks of a speaker diarization pipeline:

    * (sad) speech activity detection consists in detecting speech regions in
            an audio recording.
    * (scd) speaker change detection consists in detecting timestamps of
            speaker change point.
    * (ovl) overlapped speech detection consists in detection regions with two
            or more simultaneous speakers.
    * (emb) speaker embedding consists in projecting audio chunk into a
            (usually high-dimensional) vector space where same speaker
            embeddings are close to each other, and different speaker embeddings
            are not.
    * (dom) domain classification consists in predicting the domain of an
            audio recording

Running a complete speech activity detection experiment on the provided
"debug" dataset would go like this:

    * Run experiment on this pyannote.database protocol
      $ export DATABASE=Debug.SpeakerDiarization.Debug

    * This directory will contain experiments artifacts:
      $ mkdir my_experiment && cd my_experiment

    * A unique configuration file describes the experiment hyper-parameters
      (see "Configuration file" below for details):
      $ edit config.yml

    * This will train the model on the training set:
      $ pyannote-audio sad train ${PWD} ${DATABASE}

    * Training artifacts (including model weights) are stored in a sub-directory
      whose name makes it clear which dataset and subset (train, by default)
      were used for training the model.
      $ cd train/${DATABASE}.train

    * This will validate the model on the development set:
      $ pyannote-audio sad validate ${PWD} ${DATABASE}

    * Validation artifacts (including the selection of the best epoch) are
      stored in a sub-directory named after the dataset and subset (development,
      by default) used for validating the model.
      $ cd validate/${DATABASE}.development

    * This will apply the best model (according to the validation step) to the
      test set:
      $ pyannote-audio sad apply ${PWD} ${DATABASE}

    * Inference artifacts are stored in a sub-directory whose name makes it
      clear which epoch has been used (e.g. apply/0125). Artifacts include:
        * raw output of the best model (one numpy array per file  than can be
          loaded with pyannote.audio.features.Precomputed API and handled with
          pyannote.core.SlidingWindowFeature API)
        * (depending on the task) a file "${DATABASE}.test.rttm" containing the
          post-processing of raw output.
        * (depending on the task) a file "${DATABASE}.test.eval" containing the
          evaluation result computed with pyannote.metrics.

pyannote.database support
~~~~~~~~~~~~~~~~~~~~~~~~~

PYANNOTE_DATABASE_CONFIG=

Configuration file <root>/config.yml
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Reproducible research is facilitated by the systematic use of configuration
    files stored in <root>/config.yml in YAML format.

    .......................... <root>/config.yml ..........................
    task:
        name:
        params:

    feature_extraction:
        name:
        params:

    data_augmentation:
        name:
        params:

    architecture:
        name:
        params:

    scheduler:
        name:
        params:

    preprocessors:

    callbacks:
    ...................................................................

    File <root>/config.yml is mandatory, unless option --pretrained is used.

    When fine-tuning a model with option --pretrained=<model>, one can omit it
    and the original <model> configuration file is used instead. If (a possibly
    partial) <root>/config.yml file is provided anyway, it is used to override
    <model> configuration file.

Tensorboard support
~~~~~~~~~~~~~~~~~~~

    A bunch of metrics are logged during training and validation (e.g. loss,
    learning rate, computation time, validation metric). They can be visualized
    using tensorboard:

        $ tensorboard --logdir=<root>

Common options
~~~~~~~~~~~~~~

  <root>                  Experiment root directory. Should contain config.yml
                          configuration file, unless --pretrained option is
                          used (for which config.yml is optional).

  <protocol>              Name of protocol to use for training, validation, or
                          inference. Have a look at pyannote.database
                          documentation for instructions on how to define a
                          protocol with your own dataset:
                          https://github.com/pyannote/pyannote-database#custom-protocols

  <train>                 Path to <root> sub-directory containing training
                          artifacts (e.g. <root>/train/<protocol>.train)

  <validate>              Path to <train> sub-directory containing validation
                          artifacts (e.g. <train>/validate/<protocol>.development)
                          In case option --pretrained=<model> is used, the
                          output of the pretrained model is dumped into the
                          <validate> directory.

  --subset=<subset>       Subset to use for training (resp. validation,
                          inference). Defaults to "train" (resp. "development",
                          "test") for strict enforcement of machine learning
                          good practices.

  --gpu                   Run on GPU. When multiple GPUs are available, use
                          CUDA_VISIBLE_DEVICES environment variable to force
                          using a specific one. Defaults to using CPU if no GPU
                          is available.

  --cpu                   Run on CPU. Defaults to using GPU when available.

  --debug                 Run using PyTorch's anomaly detection. This will throw
                          an error if a NaN value is produced, and the stacktrace
                          will point to the origin of it. This option can
                          considerably slow execution.

  --from=<epoch>          Start training (resp. validating) at epoch <epoch>.
                          Use --from=last to start from last available epoch at
                          launch time. Not used for inference [default: 0].

  --to=<epoch>            End training (resp. validating) at epoch <epoch>.
                          Use --end=last to validate until last available epoch
                          at launch time. Not used for inference [default: 100].

  --batch=<size>          Set batch size used for validation and inference.
                          Has no effect when training as this parameter should
                          be defined in the configuration file [default: 32].

  --step=<ratio>          Ratio of audio chunk duration used as step between
                          two consecutive audio chunks [default: 0.25]

  --parallel=<n_jobs>     Use at most that many threads for generating training
                          samples or validating files. Defaults to using all
                          CPUs but one.


Speaker embedding
~~~~~~~~~~~~~~~~~

  --duration=<duration>   Use audio chunks with that duration. Defaults to the
                          fixed duration used during training, when available.

  --metric=<metric>       Use this metric (e.g. "cosine" or "euclidean") to
                          compare embeddings. Defaults to the metric defined in
                          <root>/config.yml configuration file.

Pretrained model options
~~~~~~~~~~~~~~~~~~~~~~~~

  --pretrained=<model>    Warm start training with pre-trained model. Can be
                          either a path to an existing checkpoint (e.g.
                          <train>/weights/0050.pt) or the name of a model
                          available in torch.hub.list('pyannote/pyannote.audio')
                          This option can also be used to apply a pretrained
                          model. See description of <validate> for more details.

Validation options
~~~~~~~~~~~~~~~~~~

  --every=<epoch>         Validate model every <epoch> epochs [default: 1].

  --evergreen             Prioritize validation of most recent epoch.

  For speech activity and overlapped speech detection, validation consists in
  looking for the value of the detection threshold that maximizes the f-score
  of recall and precision.

  For speaker change detection, validation consists in looking for the value of
  the peak detection threshold that maximizes the f-score of purity and
  coverage:

  --diarization           Use diarization purity and coverage instead of
                          (default) segmentation purity and coverage.

  For speaker embedding and verification protocols, validation runs the actual
  speaker verification experiment (representing each recording by its average
  embedding) and reports equal error rate.

  For speaker embedding and diarization protocols, validation runs a speaker
  diarization pipeline based on oracle segmentation and "pool-linkage"
  agglomerative clustering of speech turns (represented by their average
  embedding), and looks for the threshold that maximizes the f-score of purity
  and coverage.

"""

import sys
import warnings
from docopt import docopt
from pathlib import Path
import multiprocessing

import torch
from .base import apply_pretrained
from .speech_detection import SpeechActivityDetection
from .change_detection import SpeakerChangeDetection
from .overlap_detection import OverlapDetection
from .speaker_embedding import SpeakerEmbedding
from .domain_classification import DomainClassification


def main():

    # TODO: update version automatically
    arg = docopt(__doc__, version="pyannote-audio 2.0")

    params = {}

    if arg["sad"]:
        Application = SpeechActivityDetection

    elif arg["scd"]:
        Application = SpeakerChangeDetection

    elif arg["ovl"]:
        Application = OverlapDetection

    elif arg["emb"]:
        Application = SpeakerEmbedding

    elif arg["dom"]:
        Application = DomainClassification

    device = "cuda" if torch.cuda.is_available() else "cpu"
    if arg["--gpu"] and device == "cpu":
        msg = "No GPU is available. Using CPU instead."
        warnings.warn(msg)
    if arg["--cpu"] and device == "cuda":
        device = "cpu"

    params["device"] = torch.device(device)

    protocol = arg["<protocol>"]
    subset = arg["--subset"]

    if arg["--debug"]:
        msg = "Debug mode is enabled, this option might slow execution considerably."
        warnings.warn(msg, RuntimeWarning)
        torch.autograd.set_detect_anomaly(True)

    n_jobs = arg["--parallel"]
    if n_jobs is None:
        n_jobs = max(1, multiprocessing.cpu_count() - 1)
    params["n_jobs"] = int(n_jobs)

    if arg["train"]:

        params["subset"] = "train" if subset is None else subset

        # start training at this epoch (defaults to 0, but 'last' is supported)
        warm_start = arg["--from"]
        if warm_start != "last":
            warm_start = int(warm_start)

        # or start from pretrained model
        pretrained = arg["--pretrained"]
        pretrained_config_yml = None
        if pretrained is not None:

            # start from an existing model checkpoint
            # (from a different experiment)
            if Path(pretrained).exists():
                warm_start = Path(pretrained)

            else:
                try:
                    warm_start = torch.hub.load(
                        # TODO. change to 'pyannote/pyannote-audio'
                        # after 2.0 release
                        "pyannote/pyannote-audio:develop",
                        pretrained,
                    ).weights_pt_
                except Exception as e:
                    msg = (
                        f'Could not load "{warm_start}" model from torch.hub.'
                        f"The following exception was raised:\n\n{e}\n\n"
                    )
                    sys.exit(msg)

            pretrained_config_yml = warm_start.parents[3] / "config.yml"

        params["warm_start"] = warm_start

        # stop training at this epoch (defaults to never stop)
        params["epochs"] = int(arg["--to"])

        root_dir = Path(arg["<root>"]).expanduser().resolve(strict=True)
        app = Application(
            root_dir, training=True, pretrained_config_yml=pretrained_config_yml
        )
        app.train(protocol, **params)

    if arg["validate"]:

        train_dir = Path(arg["<train>"]).expanduser().resolve(strict=True)
        app = Application.from_train_dir(train_dir, training=False)

        params["subset"] = "development" if subset is None else subset

        start = arg["--from"]
        if start != "last":
            start = int(start)
        params["start"] = start

        end = arg["--to"]
        if end != "last":
            end = int(end)
        params["end"] = end

        params["every"] = int(arg["--every"])
        params["chronological"] = not arg["--evergreen"]
        params["batch_size"] = int(arg["--batch"])

        params["diarization"] = arg["--diarization"]

        duration = arg["--duration"]
        if duration is None:
            duration = getattr(app.task_, "duration", None)
            if duration is None:
                msg = (
                    "Task has no 'duration' defined. "
                    "Use '--duration' option to provide one."
                )
                raise ValueError(msg)
        else:
            duration = float(duration)
        params["duration"] = duration

        params["step"] = float(arg["--step"])

        if arg["emb"]:

            metric = arg["--metric"]
            if metric is None:
                metric = getattr(app.task_, "metric", None)
                if metric is None:
                    msg = (
                        "Approach has no 'metric' defined. "
                        "Use '--metric' option to provide one."
                    )
                    raise ValueError(msg)
            params["metric"] = metric

        # FIXME: parallel is broken in pyannote.metrics
        params["n_jobs"] = 1

        app.validate(protocol, **params)

    if arg["apply"]:

        validate_dir = Path(arg["<validate>"]).expanduser().resolve(strict=True)

        params["subset"] = "test" if subset is None else subset
        params["batch_size"] = int(arg["--batch"])

        duration = arg["--duration"]
        if duration is not None:
            duration = float(duration)
        params["duration"] = duration

        params["step"] = float(arg["--step"])
        params["Pipeline"] = getattr(Application, "Pipeline", None)

        params["pretrained"] = arg["--pretrained"]

        apply_pretrained(validate_dir, protocol, **params)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

# types
from pathlib import Path
from typing import Text, Dict, List, Tuple, Iterable
from pyannote.core import Segment

# pyannote.audio
from ..pipeline import InteractiveDiarization
from ..pipeline import PRETRAINED_PARAMS
from ..pipeline import time2index
from ..pipeline import index2index
from pyannote.audio.features import RawAudio
from pyannote.audio.features.utils import get_audio_duration

# prodigy
import prodigy
from prodigy.components.loaders import Audio
from prodigy.components.db import connect
from .utils import SAMPLE_RATE
from .utils import remove_audio_before_db
from .utils import normalize
from .utils import to_base64
from .utils import to_audio_spans

# clustering
CONSTRAINT = Tuple[float, float]
CONSTRAINTS = List[CONSTRAINT]
from pyannote.core.utils.hierarchy import propagate_constraints
from pyannote.core.utils.hierarchy import pool
from scipy.cluster.hierarchy import fcluster
from scipy.spatial.distance import cdist
from scipy.spatial.distance import pdist
from scipy.spatial.distance import squareform

# others
import numpy as np
from copy import deepcopy
from .sad import load_sad_manual
from itertools import filterfalse


class DiaRecipeHelper:
    def __init__(self, pipeline: InteractiveDiarization, dataset: Text, source: Path):
        self.pipeline = pipeline
        self.dataset = dataset
        self.source = source

    def load_dia_binary(self, path: Text):
        """Load existing examples as constraints for diarization

        This will set (or overwrite) the following attributes and return them
            * cannot_link_time
            * must_link_time
            * dont_know_time

        Parameters
        ----------
        path : Text
            Only load examples for this file.
        """

        db = connect()

        examples = [
            eg
            for eg in db.get_dataset(self.dataset)
            if eg["recipe"] == "pyannote.dia.binary" and eg["path"] == path
        ]

        cannot_link: CONSTRAINTS = [
            (eg["t1"], eg["t2"]) for eg in examples if eg["answer"] == "reject"
        ]
        must_link: CONSTRAINTS = [
            (eg["t1"], eg["t2"]) for eg in examples if eg["answer"] == "accept"
        ]
        dont_know: CONSTRAINTS = [
            (eg["t1"], eg["t2"])
            for eg in examples
            if eg["answer"] not in ["accept", "reject"]
        ]

        if len(cannot_link) > 0:
            prodigy.log(
                f"RECIPE: {path}: init: {len(cannot_link)} cannot link constraints"
            )
        if len(must_link) > 0:
            prodigy.log(f"RECIPE: {path}: init: {len(must_link)} must link constraints")

        # expand list of "cannot link" constraints thanks to the following rule
        # (u != v) & (v == w) ==> u != w
        cannot_link = propagate_constraints(cannot_link, must_link)

        self.cannot_link_time = cannot_link
        self.must_link_time = must_link
        self.dont_know_time = dont_know

    def dia_binary_stream(self) -> Iterable[Dict]:

        raw_audio = RawAudio(sample_rate=SAMPLE_RATE, mono=True)

        for audio_source in Audio(self.source):

            path = audio_source["path"]
            text = audio_source["text"]

            # load speech/non-speech annotations (from pyannote.sad.manual recipe)
            file = load_sad_manual(self.dataset, path)
            speech = file["speech"]

            if not speech:
                prodigy.log(f"RECIPE: {path}: skip: no annotated speech")
                continue

            # load existing same/different annotations (from this very recipe)
            self.load_dia_binary(path)

            # extract speaker embedding
            embedding = self.pipeline.compute_embedding(file)

            # number of consecutive steps with overlap
            window = embedding.sliding_window
            n_steps = int(np.ceil(window.duration / window.step))

            # find clean embedding (i.e. those fully included in speech regions)
            assignment = self.pipeline.get_segment_assignment(embedding, speech)
            clean = assignment > 0
            clean_embedding = embedding[clean]
            if len(clean_embedding) < 2:
                prodigy.log(f"RECIPE: {path}: skip: not enough speech")
                continue

            # conversion from "clean-only" index base to "all" index base (used later)
            clean2all = index2index(None, clean, reverse=True, return_mapping=True)

            done_with_current_file = False
            while not done_with_current_file:

                # IMPROVE do not recompute if no new constraint since last time

                # filter and convert time-based constraints in whole file referential
                # to index-based constraints in clean-only embeddings referential
                cannot_link = index2index(
                    time2index(self.cannot_link_time, window), clean
                )
                must_link = index2index(time2index(self.must_link_time, window), clean)

                dendrogram = pool(
                    clean_embedding,
                    metric="cosine",
                    cannot_link=cannot_link if cannot_link else None,
                    must_link=must_link if must_link else None,
                    must_link_method="propagate",
                )

                # # iterate from dendrogram top to bottom
                # iterations = iter(range(len(dendrogram) - 1, 0, -1))

                # iterate from merging step whose distance is the most similar
                # to the "optimal" threshold and progressively wander away from it
                iterations = filterfalse(
                    lambda i: i < 1,
                    iter(
                        np.argsort(
                            np.abs(self.pipeline.emb_threshold - dendrogram[:, 2])
                        )
                    ),
                )

                # IDEA stop annotating early once the current distance is much
                # smaller/greater than the "optimal" and we can be sure that all
                # further iterations are easy to decide.

                while True:

                    try:
                        i = next(iterations)
                    except StopIteration as e:
                        done_with_current_file = True
                        break

                    distance = dendrogram[i, 2]

                    # if distance is infinite, this is a fake clustering step
                    # prevented by a "cannot link" constraint.
                    # see pyannote.core.hierarchy.pool for details
                    if distance == np.infty:
                        prodigy.log(f"RECIPE: {path}: depth {i}: skip: cannot link")
                        continue

                    # find clusters k1 and k2 that were merged at iteration i
                    current = fcluster(
                        dendrogram, dendrogram[i, 2], criterion="distance"
                    )
                    previous = fcluster(
                        dendrogram, dendrogram[i - 1, 2], criterion="distance",
                    )
                    n_current, n_previous = max(current), max(previous)

                    # TODO handle these corner cases better
                    if n_current >= n_previous or n_previous - n_current > 1:
                        prodigy.log(f"RECIPE: {path}: depth {i}: skip: corner case")
                        continue
                    C = np.zeros((n_current, n_previous))
                    for k_current, k_previous in zip(current, previous):
                        C[k_current - 1, k_previous - 1] += 1
                    k1, k2 = (
                        np.where(C[int(np.where(np.sum(C > 0, axis=1) == 2)[0])] > 0)[0]
                        + 1
                    )

                    # find indices of embeddings fully included in clusters k1 and k2
                    neighbors1 = np.convolve(previous == k1, [1] * n_steps, mode="same")
                    indices1 = np.where(neighbors1 == n_steps)[0]
                    # if indices1.size == 0:
                    #     indices1 = np.where(neighbors1 == np.max(neighbors1))[0]

                    neighbors2 = np.convolve(previous == k2, [1] * n_steps, mode="same")
                    indices2 = np.where(neighbors2 == n_steps)[0]
                    # if indices2.size == 0:
                    #     indices2 = np.where(neighbors2 == np.max(neighbors2))[0]

                    if indices1.size == 0 or indices2.size == 0:
                        prodigy.log(
                            f"RECIPE: {path}: depth {i}: skip: too short segments"
                        )
                        continue

                    # find centroids of clusters k1 and k2
                    i1 = indices1[
                        np.argmin(
                            np.mean(
                                squareform(
                                    pdist(clean_embedding[indices1], metric="cosine")
                                ),
                                axis=1,
                            )
                        )
                    ]

                    i2 = indices2[
                        np.argmin(
                            np.mean(
                                squareform(
                                    pdist(clean_embedding[indices2], metric="cosine")
                                ),
                                axis=1,
                            )
                        )
                    ]

                    i1, i2 = sorted([i1, i2])
                    distance = cdist(
                        clean_embedding[np.newaxis, i1],
                        clean_embedding[np.newaxis, i2],
                        metric="cosine",
                    )[0, 0]

                    segment1 = window[clean2all[i1]]
                    t1 = segment1.middle
                    segment2 = window[clean2all[i2]]
                    t2 = segment2.middle

                    # did the human in the loop already provide feedback on this pair of segments?
                    pair = (t1, t2)

                    if (
                        pair in self.cannot_link_time
                        or pair in self.must_link_time
                        or pair in self.dont_know_time
                    ):
                        # do not annotate the same pair twice
                        prodigy.log(f"RECIPE: {path}: depth {i}: skip: exists")
                        continue

                    prodigy.log(f"RECIPE: {path}: depth {i}: annotate")

                    task_text = f"{text} t={t1:.1f}s vs. t={t2:.1f}s"

                    waveform1 = normalize(raw_audio.crop(file, segment1))
                    waveform2 = normalize(raw_audio.crop(file, segment2))
                    task_audio = to_base64(
                        np.vstack([waveform1, waveform2]), sample_rate=SAMPLE_RATE
                    )

                    task_audio_spans = [
                        {"start": 0.0, "end": segment1.duration, "label": "SPEAKER",},
                        {
                            "start": segment1.duration,
                            "end": segment1.duration + segment2.duration,
                            "label": "SAME_SPEAKER",
                        },
                    ]

                    yield {
                        "path": path,
                        "text": task_text,
                        "audio": task_audio,
                        "audio_spans": task_audio_spans,
                        "t1": t1,
                        "t2": t2,
                        "meta": {
                            "t1": f"{t1:.1f}s",
                            "t2": f"{t2:.1f}s",
                            "file": text,
                            "distance": f"{distance:.2f}",
                        },
                        "recipe": "pyannote.dia.binary",
                    }

                    # at that point, "dia_binary_update" is called. hence,
                    # we exit the loop because the dendrogram needs to be updated
                    break

    def dia_binary_update(self, examples: List[Dict]) -> List[Dict]:

        needs_update = False

        for eg in examples:

            t1, t2 = eg["t1"], eg["t2"]

            if eg["answer"] == "accept":
                self.must_link_time.append((t1, t2))
                needs_update = True
                prodigy.log(f"RECIPE: new constraint: +1 must link")

            elif eg["answer"] == "reject":
                self.cannot_link_time.append((t1, t2))
                needs_update = True
                prodigy.log(f"RECIPE: new constraint: +1 cannot link")

            else:
                self.dont_know_time.append((t1, t2))
                prodigy.log(f"RECIPE: new constraint: skip")

        # expand list of "cannot link" constraints thanks to the following rule
        # (u != v) & (v == w) ==> u != w
        if needs_update:
            num_cannot = len(self.cannot_link_time)
            self.cannot_link_time = propagate_constraints(
                self.cannot_link_time, self.must_link_time
            )
            new_num_cannot = len(self.cannot_link_time)
            if new_num_cannot > num_cannot:
                prodigy.log(
                    f"RECIPE: propagate constraint: +{new_num_cannot - num_cannot} cannot link"
                )

    def dia_manual_stream(self) -> Iterable[Dict]:

        for audio_source in Audio(self.source):

            path = audio_source["path"]
            text = audio_source["text"]

            # load speech/non-speech annotations (from pyannote.sad.manual recipe)
            file = load_sad_manual(self.dataset, path)
            manual_speech = file["speech"]
            annotated = file["annotated"]

            # use manual speech/non-speech annotation where available,
            # and automatic speech/non-speech else where
            duration = get_audio_duration(file)
            file_extent = Segment(0, duration)
            non_annotated = annotated.gaps(file_extent)
            if non_annotated:
                automatic_speech = self.pipeline.compute_speech(file)
                file["speech"] = automatic_speech.crop(non_annotated).update(
                    manual_speech
                )

            # load existing same/different annotations (from pyannote.dia.binary recipe)
            self.load_dia_binary(path)

            # apply speaker diarization pipeline using same/different speaker
            # binary annotation as must link/cannot link constraints
            hypothesis = self.pipeline(
                file, cannot_link=self.cannot_link_time, must_link=self.must_link_time
            )

            # rename 9 most talkative speakers to {SPEAKER_1, ..., SPEAKER_9}
            # and remaining speakers as OTHER
            mapping = {
                label: f"SPEAKER_{s+1}" if s < 9 else "OTHER"
                for s, (label, duration) in enumerate(hypothesis.chart())
            }
            hypothesis = hypothesis.rename_labels(mapping=mapping)

            audio_spans = to_audio_spans(hypothesis)
            audio_source["audio_spans"] = audio_spans
            audio_source["audio_spans_original"] = deepcopy(audio_spans)
            audio_source["recipe"] = "pyannote.dia.manual"

            yield audio_source


@prodigy.recipe(
    "pyannote.dia.binary",
    dataset=("Dataset to save annotations to", "positional", None, str),
    source=("Directory containing audio files to annotate", "positional", None, Path),
)
def dia_binary(dataset: Text, source: Path) -> Dict:

    pipeline = InteractiveDiarization().instantiate(PRETRAINED_PARAMS)
    helper = DiaRecipeHelper(pipeline, dataset, source)

    return {
        "dataset": dataset,
        "view_id": "audio",
        "stream": helper.dia_binary_stream(),
        "update": helper.dia_binary_update,
        "before_db": remove_audio_before_db,
        "config": {
            "audio_autoplay": True,
            "audio_loop": True,
            "show_audio_minimap": False,
            "audio_bar_width": 3,
            "audio_bar_height": 1,
            "labels": ["SPEAKER", "SAME_SPEAKER"],
            "batch_size": 1,
            "instant_submit": True,
        },
    }


@prodigy.recipe(
    "pyannote.dia.manual",
    dataset=("Dataset to save annotations to", "positional", None, str),
    source=("Directory containing audio files to annotate", "positional", None, Path),
)
def dia_manual(dataset: Text, source: Path) -> Dict:

    pipeline = InteractiveDiarization().instantiate(PRETRAINED_PARAMS)
    helper = DiaRecipeHelper(pipeline, dataset, source)

    return {
        "dataset": dataset,
        "view_id": "audio_manual",
        "stream": helper.dia_manual_stream(),
        "before_db": remove_audio_before_db,
        "config": {
            "audio_autoplay": True,
            "audio_loop": True,
            "show_audio_minimap": True,
            "audio_bar_width": 3,
            "audio_bar_height": 1,
            "labels": [f"SPEAKER_{i+1}" for i in range(10)] + ["OTHER"],
            "batch_size": 1,
        },
    }



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

import torch
import torch.nn as nn
from typing import Text


class Recurrent(nn.Module):
    """Recurrent layers

    Parameters
    ----------
    n_features : int
        Input feature shape.
    unit : {"LSTM", "GRU"}, optional
        Defaults to "LSTM".
    hidden_size : int, optional
        Number of features in the hidden state h. Defaults to 512.
    num_layers : int, optional
        Number of recurrent layers. Defaults to 1.
    bias : bool, optional
        If False, then the layer does not use bias weights. Defaults to True.
    dropout : float, optional
        If non-zero, introduces a Dropout layer on the outputs of each layer
        except the last layer, with dropout probability equal to dropout.
        Defaults to 0.
    bidirectional : bool, optional
        Use bidirectional RNN. Defaults to True.
    probes : bool, optional
        Split multi-layer RNN into multiple one-layer RNNs to expose
        corresponding probes (see pyannote.audio.train.model.Model.probes).
        Might be useful when using a multi-layer RNN as the trunk of a larger
        multi-task RNN tree.
    """

    def __init__(
        self,
        n_features: int,
        unit: Text = "LSTM",
        hidden_size: int = 512,
        num_layers: int = 1,
        bias: int = True,
        dropout: float = 0.0,
        bidirectional: bool = True,
        probes: bool = False,
    ):
        super().__init__()

        self.n_features = n_features
        self.unit = unit
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.bias = bias
        self.dropout = dropout
        self.bidirectional = bidirectional
        self.probes = probes

        if num_layers < 1:
            if bidirectional:
                msg = "'bidirectional' must be set to False when num_layers < 1"
                raise ValueError(msg)
            return

        Klass = getattr(nn, self.unit)

        if probes:

            self.rnn = nn.ModuleList([])

            for i in range(self.num_layers):

                if i > 0:
                    input_dim = self.hidden_size
                    if self.bidirectional:
                        input_dim *= 2
                else:
                    input_dim = self.n_features

                dropout = 0 if (i + 1 == self.num_layers) else self.dropout

                rnn = Klass(
                    input_dim,
                    self.hidden_size,
                    num_layers=1,
                    bias=self.bias,
                    batch_first=True,
                    dropout=dropout,
                    bidirectional=self.bidirectional,
                )

                self.rnn.append(rnn)

        else:
            self.rnn = Klass(
                self.n_features,
                self.hidden_size,
                num_layers=self.num_layers,
                bias=self.bias,
                batch_first=True,
                dropout=self.dropout,
                bidirectional=self.bidirectional,
            )

    def forward(self, features: torch.Tensor, **kwargs) -> torch.Tensor:
        """Apply recurrent layer

        Parameters
        ----------
        features : torch.Tensor
            Input feature sequence with shape (batch_size, n_frames, n_features).

        Returns
        -------
        output : torch.Tensor
            Output sequence with shape (batch_size, n_frames, hidden_size x n_directions).
        """

        if self.num_layers < 1:
            return features

        if self.probes:

            output, hidden = None, None
            for i, rnn in enumerate(self.rnn):

                rnn = getattr(self, f"rnn{i+1:02d}")

                if i > 0:
                    output, hidden = rnn(output, hidden)
                else:
                    output, hidden = rnn(features)

        else:
            output, hidden = self.rnn(features)

        return output

    @property
    def dimension(self):

        if self.num_layers < 1:
            return self.n_features

        dimension = self.hidden_size
        if self.bidirectional:
            dimension *= 2

        return dimension



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2016-2019 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""
Feature extraction using [`librosa`](https://librosa.github.io/librosa/)
"""

import librosa
import numpy as np

from .base import FeatureExtraction
from pyannote.core.segment import SlidingWindow


class LibrosaFeatureExtraction(FeatureExtraction):
    """librosa feature extraction base class

    Parameters
    ----------
    sample_rate : int, optional
        Defaults to 16000 (i.e. 16kHz)
    augmentation : `pyannote.audio.augmentation.Augmentation`, optional
        Data augmentation.
    duration : float, optional
        Defaults to 0.025 (25ms).
    step : float, optional
        Defaults to 0.010 (10ms).
    """

    def __init__(self, sample_rate=16000, augmentation=None, duration=0.025, step=0.01):

        super().__init__(sample_rate=sample_rate, augmentation=augmentation)
        self.duration = duration
        self.step = step

        self.sliding_window_ = SlidingWindow(
            start=-0.5 * self.duration, duration=self.duration, step=self.step
        )

    def get_resolution(self):
        return self.sliding_window_


class LibrosaSpectrogram(LibrosaFeatureExtraction):
    """librosa spectrogram

    Parameters
    ----------
    sample_rate : int, optional
        Defaults to 16000 (i.e. 16kHz)
    augmentation : `pyannote.audio.augmentation.Augmentation`, optional
        Data augmentation.
    duration : float, optional
        Defaults to 0.025.
    step : float, optional
        Defaults to 0.010.
    """

    def __init__(
        self, sample_rate=16000, augmentation=None, duration=0.025, step=0.010
    ):

        super().__init__(
            sample_rate=sample_rate,
            augmentation=augmentation,
            duration=duration,
            step=step,
        )

        self.n_fft_ = int(self.duration * self.sample_rate)
        self.hop_length_ = int(self.step * self.sample_rate)

    def get_dimension(self):
        return self.n_fft_ // 2 + 1

    def get_features(self, y, sample_rate):
        """Feature extraction

        Parameters
        ----------
        y : (n_samples, 1) numpy array
            Waveform
        sample_rate : int
            Sample rate

        Returns
        -------
        data : (n_frames, n_dimensions) numpy array
            Features
        """

        fft = librosa.core.stft(
            y=y.squeeze(),
            n_fft=self.n_fft_,
            hop_length=self.hop_length_,
            center=True,
            window="hamming",
        )
        return np.abs(fft).T


class LibrosaMelSpectrogram(LibrosaFeatureExtraction):
    """librosa mel-spectrogram

    Parameters
    ----------
    sample_rate : int, optional
        Defaults to 16000 (i.e. 16kHz)
    augmentation : `pyannote.audio.augmentation.Augmentation`, optional
        Data augmentation.
    duration : float, optional
        Defaults to 0.025.
    step : float, optional
        Defaults to 0.010.
    n_mels : int, optional
        Defaults to 96.
    """

    def __init__(
        self,
        sample_rate=16000,
        augmentation=None,
        duration=0.025,
        step=0.010,
        n_mels=96,
    ):

        super().__init__(
            sample_rate=sample_rate,
            augmentation=augmentation,
            duration=duration,
            step=step,
        )

        self.n_mels = n_mels
        self.n_fft_ = int(self.duration * self.sample_rate)
        self.hop_length_ = int(self.step * self.sample_rate)

    def get_dimension(self):
        return self.n_mels

    def get_features(self, y, sample_rate):
        """Feature extraction

        Parameters
        ----------
        y : (n_samples, 1) numpy array
            Waveform
        sample_rate : int
            Sample rate

        Returns
        -------
        data : (n_frames, n_mels) numpy array
            Features
        """

        X = librosa.feature.melspectrogram(
            y.squeeze(),
            sr=sample_rate,
            n_mels=self.n_mels,
            n_fft=self.n_fft_,
            hop_length=self.hop_length_,
            power=2.0,
        )

        return librosa.amplitude_to_db(X, ref=1.0, amin=1e-5, top_db=80.0).T


class LibrosaMFCC(LibrosaFeatureExtraction):
    """librosa MFCC

    ::

            | e    |  energy
            | c1   |
            | c2   |  coefficients
            | c3   |
            | ...  |
            | Δe   |  energy first derivative
            | Δc1  |
        x = | Δc2  |  coefficients first derivatives
            | Δc3  |
            | ...  |
            | ΔΔe  |  energy second derivative
            | ΔΔc1 |
            | ΔΔc2 |  coefficients second derivatives
            | ΔΔc3 |
            | ...  |


    Parameters
    ----------
    sample_rate : int, optional
        Defaults to 16000 (i.e. 16kHz)
    augmentation : `pyannote.audio.augmentation.Augmentation`, optional
        Data augmentation.
    duration : float, optional
        Defaults to 0.025.
    step : float, optional
        Defaults to 0.010.
    e : bool, optional
        Energy. Defaults to True.
    coefs : int, optional
        Number of coefficients. Defaults to 11.
    De : bool, optional
        Keep energy first derivative. Defaults to False.
    D : bool, optional
        Add first order derivatives. Defaults to False.
    DDe : bool, optional
        Keep energy second derivative. Defaults to False.
    DD : bool, optional
        Add second order derivatives. Defaults to False.

    Notes
    -----
    Internal setup
        * fftWindow = Hanning
        * melMaxFreq = 6854.0
        * melMinFreq = 130.0
        * melNbFilters = 40

    """

    def __init__(
        self,
        sample_rate=16000,
        augmentation=None,
        duration=0.025,
        step=0.01,
        e=False,
        De=True,
        DDe=True,
        coefs=19,
        D=True,
        DD=True,
        fmin=0.0,
        fmax=None,
        n_mels=40,
    ):

        super().__init__(
            sample_rate=sample_rate,
            augmentation=augmentation,
            duration=duration,
            step=step,
        )

        self.e = e
        self.coefs = coefs
        self.De = De
        self.DDe = DDe
        self.D = D
        self.DD = DD

        self.n_mels = n_mels  # yaafe / 40
        self.fmin = fmin  # yaafe / 130.0
        self.fmax = fmax  # yaafe / 6854.0

    def get_context_duration(self):
        return 0.0

    def get_features(self, y, sample_rate):
        """Feature extraction

        Parameters
        ----------
        y : (n_samples, 1) numpy array
            Waveform
        sample_rate : int
            Sample rate

        Returns
        -------
        data : (n_frames, n_dimensions) numpy array
            Features
        """

        # adding because C0 is the energy
        n_mfcc = self.coefs + 1

        n_fft = int(self.duration * sample_rate)
        hop_length = int(self.step * sample_rate)

        mfcc = librosa.feature.mfcc(
            y=y.squeeze(),
            sr=sample_rate,
            n_mfcc=n_mfcc,
            n_fft=n_fft,
            hop_length=hop_length,
            n_mels=self.n_mels,
            htk=True,
            fmin=self.fmin,
            fmax=self.fmax,
        )

        if self.De or self.D:
            mfcc_d = librosa.feature.delta(mfcc, width=9, order=1, axis=-1)

        if self.DDe or self.DD:
            mfcc_dd = librosa.feature.delta(mfcc, width=9, order=2, axis=-1)

        stack = []

        if self.e:
            stack.append(mfcc[0, :])

        stack.append(mfcc[1:, :])

        if self.De:
            stack.append(mfcc_d[0, :])

        if self.D:
            stack.append(mfcc_d[1:, :])

        if self.DDe:
            stack.append(mfcc_dd[0, :])

        if self.DD:
            stack.append(mfcc_dd[1:, :])

        return np.vstack(stack).T

    def get_dimension(self):
        n_features = 0
        n_features += self.e
        n_features += self.De
        n_features += self.DDe
        n_features += self.coefs
        n_features += self.coefs * self.D
        n_features += self.coefs * self.DD
        return n_features



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2019-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


from typing import Tuple
from typing import Optional
from .utils import NoiseCollection

from .base import Augmentation
from .utils import Noise
import pyroomacoustics as pra
import numpy as np
import threading
import collections

normalize = lambda wav: wav / (np.sqrt(np.mean(wav ** 2)) + 1e-8)


class Reverb(Augmentation):
    """Simulate indoor reverberation

    Parameters
    ----------
    depth : (float, float), optional
        Minimum and maximum values for room depth (in meters).
        Defaults to (2.0, 10.0).
    width : (float, float), optional
        Minimum and maximum values for room width (in meters).
        Defaults to (1.0, 10.0).
    height : (float, float), optional
        Minimum and maximum values for room heigth (in meters).
        Defaults to (2.0, 5.0).
    absorption : (float, float), optional
        Minimum and maximum values of walls absorption coefficient.
        Defaults to (0.2, 0.9).
    noise : str or list of str, optional
        `pyannote.database` collection(s) used for adding noise.
        Defaults to "MUSAN.Collection.BackgroundNoise"
    snr : (float, float), optional
        Minimum and maximum values of signal-to-noise ratio.
        Defaults to (5.0, 15.0)

    """

    def __init__(
        self,
        depth: Tuple[float, float] = (2.0, 10.0),
        width: Tuple[float, float] = (1.0, 10.0),
        height: Tuple[float, float] = (2.0, 5.0),
        absorption: Tuple[float, float] = (0.2, 0.9),
        noise: Optional[NoiseCollection] = None,
        snr: Tuple[float, float] = (5.0, 15.0),
    ):

        super().__init__()
        self.depth = depth
        self.width = width
        self.height = height
        self.absorption = absorption
        self.max_order_ = 17

        self.noise = noise
        self.snr = snr
        self.noise_ = Noise(collection=self.noise)

        self.n_rooms_ = 128
        self.new_rooms_prob_ = 0.001
        self.main_lock_ = threading.Lock()
        self.rooms_ = collections.deque(maxlen=self.n_rooms_)
        self.room_lock_ = [threading.Lock() for _ in range(self.n_rooms_)]

    @staticmethod
    def random(m: float, M: float):
        return (M - m) * np.random.random_sample() + m

    def new_room(self, sample_rate: int):

        # generate a room at random
        depth = self.random(*self.depth)
        width = self.random(*self.width)
        height = self.random(*self.height)
        absorption = self.random(*self.absorption)
        room = pra.ShoeBox(
            [depth, width, height],
            fs=sample_rate,
            absorption=absorption,
            max_order=self.max_order_,
        )

        # play the original audio chunk at a random location
        original = [
            self.random(0, depth),
            self.random(0, width),
            self.random(0, height),
        ]
        room.add_source(original)

        # play the noise audio chunk at a random location
        noise = [self.random(0, depth), self.random(0, width), self.random(0, height)]
        room.add_source(noise)

        # place the microphone at a random location
        microphone = [
            self.random(0, depth),
            self.random(0, width),
            self.random(0, height),
        ]
        room.add_microphone_array(
            pra.MicrophoneArray(np.c_[microphone, microphone], sample_rate)
        )

        room.compute_rir()

        return room

    def __call__(self, original: np.ndarray, sample_rate: int) -> np.ndarray:

        with self.main_lock_:

            # initialize rooms (with 2 sources and 1 microphone)
            while len(self.rooms_) < self.n_rooms_:
                room = self.new_room(sample_rate)
                self.rooms_.append(room)

            # create new room with probability new_rooms_prob_
            if np.random.rand() > 1.0 - self.new_rooms_prob_:
                room = self.new_room(sample_rate)
                self.rooms_.append(room)

            # choose one room at random
            index = np.random.choice(self.n_rooms_)

        # lock chosen room to ensure room.sources are not updated concurrently
        with self.room_lock_[index]:

            room = self.rooms_[index]

            # play normalized original audio chunk at source #1
            n_samples = len(original)
            original = normalize(original).squeeze()
            room.sources[0].add_signal(original)

            # generate noise with random SNR
            noise = self.noise_(n_samples, sample_rate).squeeze()
            snr = self.random(*self.snr)
            alpha = np.exp(-np.log(10) * snr / 20)
            noise *= alpha

            # play noise at source #2
            room.sources[1].add_signal(noise)

            # simulate room and return microphone signal
            room.simulate()
            return room.mic_array.signals[0, :n_samples, np.newaxis]



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2014-2018 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""
# Feature extraction
"""

from .base import FeatureExtraction

try:
    from .with_librosa import LibrosaMFCC, LibrosaSpectrogram, LibrosaMelSpectrogram
except Exception as e:
    msg = (
        f'Feature extractors based on "librosa" are not available '
        f'because something went wrong when importing them: "{e}".'
    )
    print(msg)

from .precomputed import Precomputed

try:
    from .utils import RawAudio
except Exception as e:
    msg = f"Loading raw audio might fail because something went wrong: {e}."
    print(msg)

try:
    from .pretrained import Pretrained
except Exception as e:
    msg = (
        f"Feature extraction using pretrained models are not available "
        f'because something went wrong at import: "{e}".'
    )
    print(msg)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

from typing import Optional, Iterator, Callable, Union, List

try:
    from typing import Literal
except ImportError as e:
    from typing_extensions import Literal
from pathlib import Path
import io
import os
import sys
import yaml
import torch
import tempfile
import warnings
from itertools import chain
from torch.nn import Module
from torch.optim import Optimizer, SGD
from torch.utils.tensorboard import SummaryWriter
from .logging import Logging
from .callback import Callback
from .callback import Callbacks
from .schedulers import BaseSchedulerCallback
from .schedulers import ConstantScheduler
from .generator import BatchGenerator
from .model import Model
from ..utils.timeout import timeout
from ..utils.background import AdaptiveBackgroundGenerator


ARBITRARY_LR = 0.1


class Trainer:
    """Trainer

    Attributes
    ----------
    model : Model
    specifications : dict
    device : torch.device
    """

    SPECS_YML = "{train_dir}/specs.yml"
    MODEL_PT = "{train_dir}/weights/{epoch:04d}.pt"
    OPTIMIZER_PT = "{train_dir}/weights/{epoch:04d}.optimizer.pt"

    def load_state(self, model_pt: Optional[Path] = None) -> bool:
        """Load model and optimizer states from disk

        Parameters
        ----------
        model_pt : `Path`, optional
            Path to file containing model state.
            Defaults to guessing it from trainer status.

        Returns
        -------
        success : bool
            True if state was loaded successfully, False otherwise.
        """

        if model_pt is None:
            _model_pt = self.MODEL_PT.format(
                train_dir=self.train_dir_, epoch=self.epoch_
            )
            optimizer_pt = self.OPTIMIZER_PT.format(
                train_dir=self.train_dir_, epoch=self.epoch_
            )

        else:
            _model_pt = model_pt
            optimizer_pt = model_pt.with_suffix(".optimizer.pt")

        model_state = torch.load(_model_pt, map_location=lambda storage, loc: storage)
        missing_keys, unexpected_keys = self.model_.load_state_dict(
            model_state, strict=False
        )
        if missing_keys:
            msg = f"Checkpoint misses the following weights: {missing_keys}."
            warnings.warn(msg)
        if unexpected_keys:
            msg = f"Checkpoint contains unexpected weights: {unexpected_keys}."
            warnings.warn(msg)

        success = self.load_more(model_pt=model_pt)

        if success:

            try:
                optimizer_state = torch.load(
                    optimizer_pt, map_location=lambda storage, loc: storage
                )
                self.optimizer_.load_state_dict(optimizer_state)
            except Exception as e:
                msg = (
                    f"Did not load optimizer state (most likely because current "
                    f"training session uses a different loss than the one used "
                    f"for pre-training)."
                )
                warnings.warn(msg)

        return success

    def load_more(self, model_pt=None) -> bool:
        """Called after model state is loaded

        This method can be overriden to load additional states.
        For instance, it can be used to load the state of a domain classifier
        in domain-adversarial training, or the class centers in center loss.

        Parameters
        ----------
        model_pt : `Path`, optional
            Path to file containing model state.
            Defaults to guessing it from trainer status.

        Returns
        -------
        success : bool
            True if state was loaded successfully, False otherwise.
        """
        return True

    def save_state(self):
        """Save model and optimizer states to disk"""

        # save model state
        model_pt = self.MODEL_PT.format(train_dir=self.train_dir_, epoch=self.epoch_)
        torch.save(self.model_.state_dict(), model_pt)

        # save optimizer state
        optimizer_pt = self.OPTIMIZER_PT.format(
            train_dir=self.train_dir_, epoch=self.epoch_
        )
        torch.save(self.optimizer_.state_dict(), optimizer_pt)

        self.save_more()

    def save_more(self):
        """Called after model and optimizer states are saves

        This method can be overriden to save additional states.
        For instance, it can be used to save the state of a domain classifier
        in domain-adversarial training, or the class centers in center loss.
        """
        pass

    def parameters(self):
        return chain(self.model_.parameters(), self.more_parameters())

    def more_parameters(self):
        """Called by `parameters` method

        This method can be overriden to define additional modules.
        For instance, it can be used to define a domain classifier for
        for domain-adversarial training.

        It should be an iterator yielding the parameters of these additional
        modules.

        Yields
        ------
        parameter : nn.Parameter
            Trainable trainer parameters.
        """
        return []

    def on_train_start(self):
        """Called just before training starts"""
        pass

    def on_epoch_start(self):
        """Called just before epoch starts"""
        pass

    def on_batch_start(self, batch):
        """Called just before batch is processed

        Parameters
        ----------
        batch : `dict`
            Current batch.

        Returns
        -------
        batch : `dict`
            Updated batch.
        """
        return batch

    def on_batch_end(self, loss):
        """Called just after loss is computed

        Parameters
        ----------
        loss : `dict`
            ['loss'] (`torch.Tensor`)
        """
        pass

    def on_epoch_end(self):
        """Called when epoch ends"""
        pass

    def on_train_end(self):
        """Called when training stops"""
        pass

    @property
    def model(self) -> Model:
        return self.model_

    @property
    def optimizer(self):
        return self.optimizer_

    @property
    def specifications(self):
        return self.batch_generator_.specifications

    @property
    def device(self) -> torch.device:
        return self.device_

    @property
    def epoch(self) -> int:
        return self.epoch_

    @property
    def batches_per_epoch(self) -> int:
        return self.batches_per_epoch_

    def get_new_batch(self):
        return next(self.batches_)

    def fit_iter(
        self,
        model: Model,
        batch_generator: BatchGenerator,
        warm_start: Union[int, Path] = 0,
        epochs: int = 1000,
        get_optimizer: Callable[..., Optimizer] = SGD,
        scheduler: Optional[BaseSchedulerCallback] = None,
        learning_rate: Union[Literal["auto"], float] = "auto",
        train_dir: Optional[Path] = None,
        verbosity: int = 2,
        device: Optional[torch.device] = None,
        callbacks: Optional[List[Callback]] = None,
        n_jobs: int = 1,
    ) -> Iterator[Model]:
        """Train model

        Parameters
        ----------
        model : `Model`
            Model.
        batch_generator : `BatchGenerator`
            Batch generator.
        warm_start : `int` or `Path`, optional
            Restart training at this epoch or from this model.
            Default behavior (0) is to train the model from scratch.
        epochs : `int`, optional
            Train model for that many epochs. Defaults to 1000.
        get_optimizer : `callable`, optional
            Callable taking model parameters as input and returns an instance of
            `torch.optim.Optimizer`. May also support `lr` keyword argument.
            Defaults to `torch.optim.SGD`.
        scheduler : `BaseSchedulerCallback`, optional
            Learning rate scheduler. Defaults to `ConstantScheduler`.
        learning_rate : {float, 'auto'}, optional
            Learning rate. Default behavior ('auto') is to use the AutoLR
            heuristic to determine the learning rate automatically.
        train_dir : `Path`, optional
            Directory where models and other log files are stored.
            Defaults to a temporary directory.
        verbosity : `int`, optional
            Set level of verbosity.
            Use 0 (default) to not show any progress bar.
            Use 1 to show a progress bar updated at the end of each epoch.
            Use 2 to add a second progress bar updated at the end of each batch.
        device : `torch.device`, optional
            Device on which the model will be allocated. Defaults to using CPU.
        callbacks : `list` of `Callback` instances
            Add custom callbacks.
        n_jobs : `int`, optional
            Defaults to 1.

        Yields
        ------
        model : `Model`
            Model at current iteration
        """

        #
        if train_dir is None:
            train_dir = Path(tempfile.mkdtemp())
        self.train_dir_ = train_dir

        # DEVICE
        self.device_ = torch.device("cpu") if device is None else device

        # MODEL
        self.model_ = model.to(self.device_)

        # BATCH GENERATOR
        self.batch_generator_ = batch_generator
        self.batches_ = AdaptiveBackgroundGenerator(
            self.batch_generator_, n_jobs=n_jobs
        )
        self.batches_per_epoch_ = self.batch_generator_.batches_per_epoch

        # OPTIMIZER
        lr = ARBITRARY_LR if learning_rate == "auto" else learning_rate
        self.optimizer_ = get_optimizer(self.parameters(), lr=lr)
        self.base_learning_rate_ = learning_rate

        # make sure that 'train_dir' directory does not exist when
        # fine-tuning a pre-trained model or starting from scratch
        # as it might contain the output of very long computations:
        # you do not want to erase them by mistake!

        if isinstance(warm_start, Path) or warm_start == 0:

            try:
                # this will fail if the directory already exists
                os.makedirs(self.train_dir_ / "weights")

            except FileExistsError as e:

                # ask user whether it is OK to continue
                try:
                    with timeout(60):
                        msg = (
                            f'Directory "{self.train_dir_}" exists.\n'
                            f"Are you OK to overwrite existing models? [y/N]: "
                        )
                        overwrite = (input(msg) or "n").lower()
                except TimeoutError:
                    # defaults to "no" after 60 seconds
                    overwrite = "n"

                # stop everything if the user did not say "yes" after a while
                if overwrite != "y":
                    sys.exit()

        # defaults to 0
        self.epoch_ = 0

        # when warm_start is an integer, it means that the user wants to
        # restart training at a given epoch. we intialize the model state and
        # set epoch_ attribute accordingly.
        if isinstance(warm_start, int):

            if warm_start > 0:

                # set epoch_ to requested value...
                self.epoch_ = warm_start

                # ... and load corresponding model if requested
                success = self.load_state(model_pt=None)

        # when warm_start is a Path, it means that the user wants to
        # restart training from a pretrained model
        else:
            try:
                success = self.load_state(model_pt=warm_start)

            except Exception as e:
                msg = (
                    f"Could not assign model weights. The following exception "
                    f"was raised:\n\n{e}\n\nAre you sure the architectures "
                    f"are consistent?"
                )
                sys.exit(msg)

            # save pretrained model as epoch 0
            self.save_state()

        # save specifications to weights/specs.yml
        specs_yml = self.SPECS_YML.format(train_dir=self.train_dir_)
        with io.open(specs_yml, "w") as fp:
            specifications = dict(self.specifications)
            specifications["task"] = str(specifications["task"])
            yaml.dump(specifications, fp, default_flow_style=False)

        # TODO in case success = False, one should freeze the main network for
        # TODO a few epochs and train the "more" part alone first before
        # TODO unfreezing everything. this could be done through a callback

        callbacks_ = []

        # SCHEDULER
        if scheduler is None:
            scheduler = ConstantScheduler()
        callbacks_.append(scheduler)

        logger = Logging(epochs=epochs, verbosity=verbosity)
        callbacks_.append(logger)

        # CUSTOM CALLBACKS
        if callbacks is not None:
            callbacks_.extend(callbacks)

        callbacks = Callbacks(callbacks_)

        # TRAINING STARTS
        self.tensorboard_ = SummaryWriter(
            log_dir=self.train_dir_, purge_step=self.epoch_
        )
        self.on_train_start()
        callbacks.on_train_start(self)

        while self.epoch_ < epochs:

            # EPOCH STARTS
            self.epoch_ += 1
            self.on_epoch_start()
            callbacks.on_epoch_start(self)

            for i in range(self.batches_per_epoch_):

                batch = self.get_new_batch()

                # BATCH IS READY FOR FORWARD PASS
                batch = self.on_batch_start(batch)
                batch = callbacks.on_batch_start(self, batch)

                # FORWARD PASS + LOSS COMPUTATION
                loss = self.batch_loss(batch)

                # BACKWARD PASS
                loss["loss"].backward()
                self.optimizer_.step()
                self.optimizer_.zero_grad()

                # OPTIMIZATION STEP IS DONE
                self.on_batch_end(loss)
                callbacks.on_batch_end(self, loss)

            self.on_epoch_end()
            callbacks.on_epoch_end(self)

            yield self.model_

            self.save_state()

        callbacks.on_train_end(self)

        self.batches_.deactivate()



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2019 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Juan Manuel Coria

# The TDNN class defined here was taken from https://github.com/jonasvdd/TDNN

# Please give proper credit to the authors if you are using TDNN based or X-Vector based
# models by citing their papers:

# Waibel, Alexander H., Toshiyuki Hanazawa, Geoffrey E. Hinton, Kiyohiro Shikano and Kevin J. Lang.
# "Phoneme recognition using time-delay neural networks."
# IEEE Trans. Acoustics, Speech, and Signal Processing 37 (1989): 328-339.
# https://pdfs.semanticscholar.org/cd62/c9976534a6a2096a38244f6cbb03635a127e.pdf?_ga=2.86820248.1800960571.1579515113-23298545.1575886658

# Peddinti, Vijayaditya, Daniel Povey and Sanjeev Khudanpur.
# "A time delay neural network architecture for efficient modeling of long temporal contexts."
# INTERSPEECH (2015).
# https://www.danielpovey.com/files/2015_interspeech_multisplice.pdf

# Snyder, David, Daniel Garcia-Romero, Gregory Sell, Daniel Povey and Sanjeev Khudanpur.
# "X-Vectors: Robust DNN Embeddings for Speaker Recognition."
# 2018 IEEE International Conference on Acoustics, Speech and Signal Processing (ICASSP) (2018): 5329-5333.
# https://www.danielpovey.com/files/2018_icassp_xvectors.pdf

from typing import Optional

import torch
import torch.nn as nn
from torch.nn.utils import weight_norm
import torch.nn.functional as F

from .pooling import StatsPool


class TDNN(nn.Module):
    def __init__(
        self,
        context: list,
        input_channels: int,
        output_channels: int,
        full_context: bool = True,
    ):
        """
        Implementation of a 'Fast' TDNN layer by exploiting the dilation argument of the PyTorch Conv1d class

        Due to its fastness the context has gained two constraints:
            * The context must be symmetric
            * The context must have equal spacing between each consecutive element

        For example: the non-full and symmetric context {-3, -2, 0, +2, +3} is not valid since it doesn't have
        equal spacing; The non-full context {-6, -3, 0, 3, 6} is both symmetric and has an equal spacing, this is
        considered valid.

        :param context: The temporal context
        :param input_channels: The number of input channels
        :param output_channels: The number of channels produced by the temporal convolution
        :param full_context: Indicates whether a full context needs to be used
        """
        super(TDNN, self).__init__()
        self.full_context = full_context
        self.input_dim = input_channels
        self.output_dim = output_channels

        context = sorted(context)
        self.check_valid_context(context, full_context)

        if full_context:
            kernel_size = context[-1] - context[0] + 1 if len(context) > 1 else 1
            self.temporal_conv = weight_norm(
                nn.Conv1d(input_channels, output_channels, kernel_size)
            )
        else:
            # use dilation
            delta = context[1] - context[0]
            self.temporal_conv = weight_norm(
                nn.Conv1d(
                    input_channels,
                    output_channels,
                    kernel_size=len(context),
                    dilation=delta,
                )
            )

    def forward(self, x):
        """
        :param x: is one batch of data, x.size(): [batch_size, sequence_length, input_channels]
            sequence length is the dimension of the arbitrary length data
        :return: [batch_size, len(valid_steps), output_dim]
        """
        x = self.temporal_conv(torch.transpose(x, 1, 2))
        return F.relu(torch.transpose(x, 1, 2))

    @staticmethod
    def check_valid_context(context: list, full_context: bool) -> None:
        """
        Check whether the context is symmetrical and whether and whether the passed
        context can be used for creating a convolution kernel with dil

        :param full_context: indicates whether the full context (dilation=1) will be used
        :param context: The context of the model, must be symmetric if no full context and have an equal spacing.
        """
        if full_context:
            assert (
                len(context) <= 2
            ), "If the full context is given one must only define the smallest and largest"
            if len(context) == 2:
                assert context[0] + context[-1] == 0, "The context must be symmetric"
        else:
            assert len(context) % 2 != 0, "The context size must be odd"
            assert (
                context[len(context) // 2] == 0
            ), "The context contain 0 in the center"
            if len(context) > 1:
                delta = [context[i] - context[i - 1] for i in range(1, len(context))]
                assert all(
                    delta[0] == delta[i] for i in range(1, len(delta))
                ), "Intra context spacing must be equal!"


class XVectorNet(nn.Module):
    """
    X-Vector neural network architecture as defined by https://www.danielpovey.com/files/2018_icassp_xvectors.pdf

    Parameters
    ----------
    input_dim : int, default 24
        dimension of the input frames
    embedding_dim : int, default 512
        dimension of latent embeddings
    """

    @property
    def dimension(self):
        return self.embedding_dim

    def __init__(self, input_dim: int = 24, embedding_dim: int = 512):
        super(XVectorNet, self).__init__()
        frame1 = TDNN(
            context=[-2, 2],
            input_channels=input_dim,
            output_channels=512,
            full_context=True,
        )
        frame2 = TDNN(
            context=[-2, 0, 2],
            input_channels=512,
            output_channels=512,
            full_context=False,
        )
        frame3 = TDNN(
            context=[-3, 0, 3],
            input_channels=512,
            output_channels=512,
            full_context=False,
        )
        frame4 = TDNN(
            context=[0], input_channels=512, output_channels=512, full_context=True
        )
        frame5 = TDNN(
            context=[0], input_channels=512, output_channels=1500, full_context=True
        )
        self.tdnn = nn.Sequential(frame1, frame2, frame3, frame4, frame5, StatsPool())
        self.segment6 = nn.Linear(3000, embedding_dim)
        self.segment7 = nn.Linear(embedding_dim, embedding_dim)
        self.embedding_dim = embedding_dim

    def forward(self, x: torch.Tensor, return_intermediate: Optional[str] = None):
        """Calculate X-Vector network activations.
           Return the requested intermediate layer without computing unnecessary activations.

        Parameters
        ----------
        x : (batch_size, n_frames, out_channels)
            Batch of frames
        return_intermediate : 'stats_pool' | 'segment6' | 'segment7' | None
            If specified, return the activation of this specific layer.
            segment6 and segment7 activations are returned before the application of non linearity.

        Returns
        -------
        activations :
            (batch_size, 3000)               if return_intermediate == 'stats_pool'
            (batch_size, embedding_dim)      if return_intermediate == 'segment6' | 'segment7' | None
        """

        x = self.tdnn(x)

        if return_intermediate == "stats_pool":
            return x

        x = self.segment6(x)

        if return_intermediate == "segment6":
            return x

        x = self.segment7(F.relu(x))

        if return_intermediate == "segment7":
            return x

        return F.relu(x)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2019 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

import time
import numpy as np
from tqdm import tqdm
from .callback import Callback


class Logging(Callback):
    """Log loss and processing time to tensorboard and progress bar

    Parameters
    ----------
    epochs : `int`, optional
        Total number of epochs. Main progress bar will be prettier if provided.
    verbosity : `int`, optional
        Set level of verbosity.
        Use 0 (default) to not show any progress bar.
        Use 1 to show a progress bar updated at the end of each epoch.
        Use 2 to add a second progress bar updated at the end of each batch.
    """

    def __init__(self, epochs: int = None, verbosity: int = 0):
        super().__init__()
        self.epochs = epochs
        self.verbosity = verbosity
        self.beta_ = 0.98

    def on_train_start(self, trainer):
        if self.verbosity > 0:
            self.epochs_pbar_ = tqdm(
                desc=f"Training",
                total=self.epochs,
                leave=True,
                ncols=80,
                unit="epoch",
                initial=trainer.epoch_,
                position=0,
            )

    def on_epoch_start(self, trainer):

        # time spent in batch generation
        self.t_batch_ = list()
        # time spent in forward/backward
        self.t_model_ = list()

        # loss moving average
        self.n_batches_ = 0
        self.loss_moving_avg_ = dict()

        if self.verbosity > 0:
            self.epochs_pbar_.update(1)

        if self.verbosity > 1:
            self.batches_pbar_ = tqdm(
                desc=f"Epoch #{trainer.epoch_}",
                total=trainer.batches_per_epoch_,
                leave=False,
                ncols=80,
                unit="batch",
                position=1,
            )

        self.t_batch_end_ = time.time()

    def on_batch_start(self, trainer, batch):

        # mark time just before forward/backward
        self.t_batch_start_ = time.time()

        # time spent in batch generation
        self.t_batch_.append(self.t_batch_start_ - self.t_batch_end_)

        return batch

    def on_batch_end(self, trainer, batch_loss):

        # mark time just after forward/backward
        self.t_batch_end_ = time.time()

        # time spent in forward/backward
        self.t_model_.append(self.t_batch_end_ - self.t_batch_start_)

        self.n_batches_ += 1

        self.loss = dict()
        for key in batch_loss:
            if not key.startswith("loss"):
                continue
            loss = batch_loss[key].detach().cpu().item()
            self.loss_moving_avg_[key] = (
                self.beta_ * self.loss_moving_avg_.setdefault(key, 0.0)
                + (1 - self.beta_) * loss
            )
            self.loss[key] = self.loss_moving_avg_[key] / (
                1 - self.beta_ ** self.n_batches_
            )

        if self.verbosity > 1:
            self.batches_pbar_.set_postfix(ordered_dict=self.loss)
            self.batches_pbar_.update(1)

    def on_epoch_end(self, trainer):

        for key, loss in self.loss.items():
            trainer.tensorboard_.add_scalar(
                f"train/{key}", loss, global_step=trainer.epoch_
            )

        trainer.tensorboard_.add_histogram(
            "profiling/model",
            np.array(self.t_model_),
            global_step=trainer.epoch_,
            bins="fd",
        )

        trainer.tensorboard_.add_histogram(
            "profiling/batch",
            np.array(self.t_batch_),
            global_step=trainer.epoch_,
            bins="fd",
        )



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr


import math


def get_conv1d_output_shape(input_shape, kernel_size, stride=1, padding=0, dilation=1):
    """Predict output shape of Conv1D"""

    out_shape = input_shape + 2 * padding - dilation * (kernel_size - 1) - 1
    out_shape = out_shape / stride + 1
    return int(math.floor(out_shape))


def get_conv2d_output_shape(input_shape, kernel_size, stride=1, padding=0, dilation=1):
    """Predict output shape of Conv2D"""

    if not isinstance(kernel_size, tuple):
        kernel_size = (kernel_size, kernel_size)
    if not isinstance(stride, tuple):
        stride = (stride, stride)
    if not isinstance(padding, tuple):
        padding = (padding, padding)
    if not isinstance(dilation, tuple):
        dilation = (dilation, dilation)

    h_in, w_in = input_shape

    h_out = h_in + 2 * padding[0] - dilation[0] * (kernel_size[0] - 1) - 1
    h_out = h_out / stride[0] + 1

    w_out = w_in + 2 * padding[1] - dilation[1] * (kernel_size[1] - 1) - 1
    w_out = w_out / stride[1] + 1

    return int(math.floor(h_out)), int(math.floor(w_out))



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2017-2019 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Ruiqing YIN - yin@limsi.fr
# Hervé BREDIN - http://herve.niderb.fr

from functools import partial
import scipy.optimize
from .base_labeling import BaseLabeling
from pyannote.database import get_annotated

from pyannote.audio.features import Pretrained
from pyannote.audio.pipeline.speaker_change_detection import (
    SpeakerChangeDetection as SpeakerChangeDetectionPipeline,
)


def validate_helper_func(current_file, pipeline=None, metric=None):
    reference = current_file["annotation"]
    uem = get_annotated(current_file)
    hypothesis = pipeline(current_file)
    return metric(reference, hypothesis, uem=uem)


class SpeakerChangeDetection(BaseLabeling):

    Pipeline = SpeakerChangeDetectionPipeline

    def validation_criterion(self, protocol, diarization=False, **kwargs):
        if diarization:
            return "diarization_fscore"
        else:
            return "segmentation_fscore"

    def validate_epoch(
        self,
        epoch,
        validation_data,
        device=None,
        batch_size=32,
        diarization=False,
        n_jobs=1,
        duration=None,
        step=0.25,
        **kwargs
    ):

        # compute (and store) SCD scores
        pretrained = Pretrained(
            validate_dir=self.validate_dir_,
            epoch=epoch,
            duration=duration,
            step=step,
            batch_size=batch_size,
            device=device,
        )

        for current_file in validation_data:
            current_file["scores"] = pretrained(current_file)

        # pipeline
        pipeline = self.Pipeline(scores="@scores", fscore=True, diarization=diarization)

        def fun(threshold):
            pipeline.instantiate({"alpha": threshold, "min_duration": 0.100})
            metric = pipeline.get_metric(parallel=True)
            validate = partial(validate_helper_func, pipeline=pipeline, metric=metric)
            if n_jobs > 1:
                _ = self.pool_.map(validate, validation_data)
            else:
                for file in validation_data:
                    _ = validate(file)

            return 1.0 - abs(metric)

        res = scipy.optimize.minimize_scalar(
            fun, bounds=(0.0, 1.0), method="bounded", options={"maxiter": 10}
        )

        threshold = res.x.item()

        return {
            "metric": self.validation_criterion(None, diarization=diarization),
            "minimize": False,
            "value": float(1.0 - res.fun),
            "pipeline": pipeline.instantiate(
                {"alpha": threshold, "min_duration": 0.100}
            ),
        }



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright © 2019 Jan Freyberg

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import torch.nn as nn
from torch.autograd import Function


class RevGrad(Function):
    @staticmethod
    def forward(ctx, input_):
        ctx.save_for_backward(input_)
        output = input_
        return output

    @staticmethod
    def backward(ctx, grad_output):  # pragma: no cover
        grad_input = None
        if ctx.needs_input_grad[0]:
            grad_input = -grad_output
        return grad_input


revgrad = RevGrad.apply


class GradientReversal(nn.Module):
    def __init__(self, *args, **kwargs):
        """
        A gradient reversal layer.
        This layer has no parameters, and simply reverses the gradient
        in the backward pass.
        """

        super().__init__(*args, **kwargs)

    def forward(self, input_):
        return revgrad(input_)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2018-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

from typing import Optional
from typing import Union
import numpy as np

from pyannote.pipeline import Pipeline
from pyannote.pipeline.parameter import Uniform

from pyannote.core import Annotation
from pyannote.core import SlidingWindowFeature

from pyannote.audio.utils.signal import Peak
from pyannote.audio.features import Precomputed

from pyannote.database import get_annotated
from pyannote.database import get_unique_identifier
from pyannote.metrics.segmentation import SegmentationPurityCoverageFMeasure
from pyannote.metrics.diarization import DiarizationPurityCoverageFMeasure

from pyannote.audio.features.wrapper import Wrapper, Wrappable


class SpeakerChangeDetection(Pipeline):
    """Speaker change detection pipeline

    Parameters
    ----------
    scores : Wrappable, optional
        Describes how raw speaker change detection scores should be obtained.
        See pyannote.audio.features.wrapper.Wrapper documentation for details.
        Defaults to "@scd_scores" that indicates that protocol files provide
        the scores in the "scd_scores" key.
    purity : `float`, optional
        Target segments purity. Defaults to 0.95.
    fscore : bool, optional
        Optimize (precision/recall) fscore. Defaults to optimizing coverage at
        given target `purity`.
    diarization : bool, optional
        Use diarization purity and coverage. Defaults to segmentation purity
        and coverage.

    Hyper-parameters
    ----------------
    alpha : `float`
        Peak detection threshold.
    min_duration : `float`
        Segment minimum duration.
    """

    def __init__(
        self,
        scores: Wrappable = None,
        purity: Optional[float] = 0.95,
        fscore: bool = False,
        diarization: bool = False,
    ):
        super().__init__()

        if scores is None:
            scores = "@scd_scores"
        self.scores = scores
        self._scores = Wrapper(self.scores)

        self.purity = purity
        self.fscore = fscore
        self.diarization = diarization

        # hyper-parameters
        self.alpha = Uniform(0.0, 1.0)
        self.min_duration = Uniform(0.0, 10.0)

    def initialize(self):
        """Initialize pipeline with current set of parameters"""

        self._peak = Peak(alpha=self.alpha, min_duration=self.min_duration)

    def __call__(self, current_file: dict) -> Annotation:
        """Apply change detection

        Parameters
        ----------
        current_file : `dict`
            File as provided by a pyannote.database protocol.  May contain a
            'scd_scores' key providing precomputed scores.

        Returns
        -------
        speech : `pyannote.core.Annotation`
            Speech regions.
        """

        scd_scores = self._scores(current_file)

        # if this check has not been done yet, do it once and for all
        if not hasattr(self, "log_scale_"):
            # heuristic to determine whether scores are log-scaled
            if np.nanmean(scd_scores.data) < 0:
                self.log_scale_ = True
            else:
                self.log_scale_ = False

        data = np.exp(scd_scores.data) if self.log_scale_ else scd_scores.data

        # take the final dimension
        # (in order to support both classification, multi-class classification,
        # and regression scores)
        change_prob = SlidingWindowFeature(data[:, -1], scd_scores.sliding_window)

        # peak detection
        change = self._peak.apply(change_prob)
        change.uri = current_file.get("uri", None)

        return change.to_annotation(generator="string", modality="audio")

    def get_metric(
        self, parallel=False
    ) -> Union[DiarizationPurityCoverageFMeasure, SegmentationPurityCoverageFMeasure]:
        """Return new instance of f-score metric"""

        if not self.fscore:
            raise NotImplementedError()

        if self.diarization:
            return DiarizationPurityCoverageFMeasure(parallel=parallel)

        return SegmentationPurityCoverageFMeasure(tolerance=0.5, parallel=parallel)

    def loss(self, current_file: dict, hypothesis: Annotation) -> float:
        """Compute (1 - coverage) at target purity

        If purity < target, return 1 + (1 - purity)

        Parameters
        ----------
        current_file : `dict`
            File as provided by a pyannote.database protocol.
        hypothesis : `pyannote.core.Annotation`
            Speech regions.

        Returns
        -------
        error : `float`
            1. - segment coverage.
        """

        metric = SegmentationPurityCoverageFMeasure(tolerance=0.500, beta=1)
        reference = current_file["annotation"]
        uem = get_annotated(current_file)
        f_measure = metric(reference, hypothesis, uem=uem)
        purity, coverage, _ = metric.compute_metrics()
        if purity > self.purity:
            return 1.0 - coverage
        else:
            return 1.0 + (1.0 - purity)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

from typing import Text, List, Dict, Iterator
from pyannote.core import Segment, SlidingWindow, Annotation

import numpy as np
import random

import io
import base64
import scipy.io.wavfile

SAMPLE_RATE = 16000


def normalize(waveform: np.ndarray) -> np.ndarray:
    """Normalize waveform for better display in Prodigy UI"""
    return waveform / (np.max(np.abs(waveform)) + 1e-8)


def to_base64(waveform: np.ndarray, sample_rate: int = SAMPLE_RATE) -> Text:
    """Convert waveform to base64 data"""
    with io.BytesIO() as content:
        scipy.io.wavfile.write(content, sample_rate, waveform)
        content.seek(0)
        b64 = base64.b64encode(content.read()).decode()
        b64 = f"data:audio/x-wav;base64,{b64}"
    return b64


def to_audio_spans(annotation: Annotation, focus: Segment = None) -> Dict:
    """Convert pyannote.core.Annotation to Prodigy's audio_spans

    Parameters
    ----------
    annotation : Annotation
        Annotation with t=0s time origin.
    focus : Segment, optional
        When provided, use its start time as audio_spans time origin.

    Returns
    -------
    audio_spans : list of dict
    """
    shift = 0.0 if focus is None else focus.start
    return [
        {"start": segment.start - shift, "end": segment.end - shift, "label": label}
        for segment, _, label in annotation.itertracks(yield_label=True)
    ]


def remove_audio_before_db(examples: List[Dict]) -> List[Dict]:
    """Remove (potentially heavy) 'audio' key from examples

    Parameters
    ----------
    examples : list of dict
        Examples.

    Returns
    -------
    examples : list of dict
        Examples with 'audio' key removed.
    """
    for eg in examples:
        if "audio" in eg:
            del eg["audio"]

    return examples


def chunks(
    duration: float, chunk: float = 30, shuffle: bool = False
) -> Iterator[Segment]:
    """Partition [0, duration] time range into smaller chunks

    Parameters
    ----------
    duration : float
        Total duration, in seconds.
    chunk : float, optional
        Chunk duration, in seconds. Defaults to 30.
    shuffle : bool, optional
        Yield chunks in random order. Defaults to chronological order.

    Yields
    ------
    focus : Segment
    """

    sliding_window = SlidingWindow(start=0.0, step=chunk, duration=chunk)
    whole = Segment(0, duration)

    if shuffle:
        chunks_ = list(chunks(duration, chunk=chunk, shuffle=False))
        random.shuffle(chunks_)
        for chunk in chunks_:
            yield chunk

    else:
        for window in sliding_window(whole):
            yield window
        if window.end < duration:
            yield Segment(window.end, duration)



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2016-2020 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""
`pyannote.audio` provides

  * speech activity detection
  * speaker change detection
  * speaker embedding
  * speaker diarization pipeline

## Installation

```bash
$ pip install pyannote.audio
```

## Citation

If you use `pyannote.audio` please use the following citations.

  - Speech  activity and speaker change detection

        @inproceedings{Yin2017,
          Author = {Ruiqing Yin and Herv\'e Bredin and Claude Barras},
          Title = {{Speaker Change Detection in Broadcast TV using Bidirectional Long Short-Term Memory Networks}},
          Booktitle = {{18th Annual Conference of the International Speech Communication Association, Interspeech 2017}},
          Year = {2017},
          Month = {August},
          Address = {Stockholm, Sweden},
          Url = {https://github.com/yinruiqing/change_detection}
        }

  - Speaker embedding

        @inproceedings{Bredin2017,
            author = {Herv\'{e} Bredin},
            title = {{TristouNet: Triplet Loss for Speaker Turn Embedding}},
            booktitle = {42nd IEEE International Conference on Acoustics, Speech and Signal Processing, ICASSP 2017},
            year = {2017},
            url = {http://arxiv.org/abs/1609.04301},
        }

  - Speaker diarization pipeline

        @inproceedings{Yin2018,
          Author = {Ruiqing Yin and Herv\'e Bredin and Claude Barras},
          Title = {{Neural Speech Turn Segmentation and Affinity Propagation for Speaker Diarization}},
          Booktitle = {{19th Annual Conference of the International Speech Communication Association, Interspeech 2018}},
          Year = {2018},
          Month = {September},
          Address = {Hyderabad, India},
        }

"""

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions
